#!/usr/bin/env python

""" metano Flux Variability Analyzer

This script performs flux variability analysis on the FBA solution for a
metabolic network (given as a set of reaction equations), an objective
function, and a number of inequality constraints.

It first performs a standard FBA (or reads in an FBA solution) and then performs
variability analysis by the following algorithm:

- Introduce a new constraint: Original objective function must have a value
                              of at least 95% of the optimal value (as
                              determined from the original FBA; threshold is
                              configurable)
- Perform LP for each flux variable: Successively maximize and minimize each of
  the flux variables


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function
from __future__ import absolute_import

from builtins import map
from builtins import range
from builtins import object
import os
import csv
from metano.defines import FbaParam, padNumber, COPYRIGHT_VERSION_STRING
from metano.fba import OptionParser, FbAnalyzer
from metano.reactionparser import ReactionParser
from metano.paramparser import ParamParser
from metano.metabolicmodel import MetabolicModel
from metano.metabolicflux import MetabolicFlux
from metano.linearproblem import LinearProblem
from numpy import array, vstack, append, dot, nan, isinf


EPSILON = 1E-9
LARGE_NUM = 1E5


class FvAnalyzer(object):
    """ Class for flux variability analysis
    """

    def __init__(self, solver="default"):
        """ initialize FVA class

        Keyword arguments:

        solver    -- LP solver to be used
        """
        self.solver = solver

    def run(self, objective, matrix, lb, ub, threshold=.95, eqs=[], ineqs=[]):
        """ successively formulate and solve linear problems for each flux with
            inequality constraint 'objective value <= threshold'

        Keyword arguments:

        objective -- coefficient vector for linear objective function
        matrix    -- stoichiometric matrix
        lb        -- list of lower bounds, indexed like reactions
        ub        -- list of upper bounds, indexed like reactions
        threshold -- objective function threshold
        eqs       -- list of (coefficient vector, right-hand side) pairs for
                     additional equality constraints
        ineqs     -- list of - '' - for additional inequality constraints

        Returns:  list of pairs (minimum, maximum), indexed like reactions
        """
        # Construct matrices of original equality and inequality constraints
        Aeq, beq, Aineq, bineq = FbAnalyzer.makeConstraintMatrices(matrix, eqs,
                                                                   ineqs)
        #lb = [-1000]*939
        #ub = [1000]*939
        # Combine original inequality constraints and
        # 'objective function <= threshold'
        if Aineq is None:
            Aineq = [objective]
            bineq = [threshold]
        else:
            Aineq = vstack((Aineq, [objective]))
            bineq = append(bineq, threshold)

        nReactions = len(lb)
        lbVec, ubVec = [], []
        for i in range(nReactions):
            if isinf(lb[i]):
                lbVec.append(-LARGE_NUM)
            else:
                lbVec.append(lb[i])
            if isinf(ub[i]):
                ubVec.append(LARGE_NUM)
            else:
                ubVec.append(ub[i])

        ps = LinearProblem(Aeq, beq, Aineq, bineq, lbVec, ubVec, self.solver)
        minmax = [None]*nReactions

        # First maximize each flux
        for index in range(nReactions):

            ps.setObjective([0.]*index+[-1.]+[0.]*(nReactions-index-1))
            s = ps.minimize()[1]
            ps.resetObjective()

            # catch random infeasible solution that resolves if the solver is recreated
            if len(s) == 0:
                ps = LinearProblem(Aeq, beq, Aineq, bineq,
                                   lbVec, ubVec, self.solver)
                ps.setObjective([0.]*index+[-1.]+[0.]*(nReactions-index-1))
                s = ps.minimize()[1]
                ps.resetObjective()

            if len(s) != 0:
                maxval = s[index]
            else:
                maxval = nan
                s = None
            minmax[index] = maxval

        # Now minimize each flux
        for index in range(nReactions):

            ps.setObjective([0.]*index+[1.]+[0.]*(nReactions-index-1))
            s = ps.minimize()[1]
            ps.resetObjective()

            # catch random infeasible solution that resolves if the solver is recreated
            if len(s) == 0:
                ps = LinearProblem(Aeq, beq, Aineq, bineq,
                                   lbVec, ubVec, self.solver)
                ps.setObjective([0.]*index+[-1.]+[0.]*(nReactions-index-1))
                s = ps.minimize()[1]
                ps.resetObjective()

            if len(s) != 0:
                minval = s[index]
            else:
                minval = nan
                s = None
            minmax[index] = minval, minmax[index]

        return minmax

    def runOnModel(self, model, fbaParams, threshp=.95, solution=None,
                   splitFluxes=True, rmDeadEnds=True):
        """ perform flux variability analysis on the given model with
            objective value <= threshp * maximum

        Keyword arguments:

        model        -- the MetabolicModel
        fbaParams    -- FBA parameters
        threshp      -- threshold percentage (objective_value <= thresp*maximum)
        solution     -- optional: either FBA solution file or solution as
                        MetabolicFlux object
        splitFluxes  -- if True, run split fluxes (resulting in non-negative
                        flux variables) before FBA
        rmDeadEnds   -- if True, remove all reactions with dead ends before
                        analysis (faster and gives an optimal solution, as well)

        Returns:
        minmax, sFlux, lb, ub, dimReduced

        minmax       -- list of pairs (minimum, maximum), indexed like reactions
        sFlux        -- FBA solution (vector indexed like reactions)
        lb, ub       -- lower/upper bounds vectors, indexed like reactions
        dimReduced   -- pair (nRows, nColumns) with dimensions of reduced matrix
        """
        if rmDeadEnds:
            deadReactions = model.findDeadEnds(True)[1]
            modelRed = model.getSubModelByExcludeList(deadReactions)
            cbz = model.canBeZero(deadReactions)
            nonZeroDeadEnds = [deadReactions[i] for i in
                               range(len(deadReactions)) if not cbz[i]]
            if nonZeroDeadEnds:
                print ("The following blocked reactions are constrained to a "
                       "non-zero flux:\n  " + "\n  ".join(nonZeroDeadEnds) +
                       "\nThe problem is infeasible.")
                lbFull, ubFull = list(map(array, model.getBounds()))
                return ([], MetabolicFlux(), lbFull, ubFull,
                        array(modelRed.getStoichiometricMatrix()).shape)
        else:
            modelRed = model

        matrix = array(modelRed.getStoichiometricMatrix())
        dimReduced = matrix.shape
        lb, ub = list(map(array, modelRed.getBounds()))

        # Build (negative) objective function vector
        objective = ParamParser.convertObjFuncToLinVec(fbaParams.objStr,
                                                       modelRed.reactionDict, -1, fbaParams.maxmin)
        maxmin_factor = -1. if fbaParams.maxmin else 1.

        # Case 1. If an FBA solution is given, use that

        if isinstance(solution, MetabolicFlux):
            # Sort solution like matrix columns
            sFlux = solution.getVecOrderedByModel(modelRed)
            # Evaluate the objective function at solution
            obj_value = maxmin_factor*dot(objective, sFlux)

        # Case 2. If solution file is given, read solution

        elif solution is not None:
            sFlux = MetabolicFlux()
            try:
                sFlux.readFromFile(solution)
            except IOError as strerror:
                print ("An error occurred while trying to read file %s:" %
                       os.path.basename(solution))
                print(strerror)
                exit()
            except SyntaxError as strerror:
                print ("An error occurred parsing file %s:" %
                       os.path.basename(solution))
                print(strerror)
                exit()

            if not sFlux.hasSameReactions(model):
                print("Error: Solution and model must have the same reactions.")
                exit()

            # Sort solution like matrix columns
            sFlux = sFlux.getVecOrderedByModel(modelRed)

            # Evaluate the objective function at solution
            obj_value = maxmin_factor*dot(objective, sFlux)

        # Case 3. Perform flux balance analysis on the metabolic network
        #         (only if no solution file is given)

        else:
            fba = FbAnalyzer(self.solver)

            if splitFluxes:
                # Split flux variables for FBA
                matrixSplit, reactionsSplit, lbSplit, ubSplit = \
                    FbAnalyzer.splitFluxes(matrix, modelRed.getReactionNames(),
                                           lb, ub)
                obj_value, sFlux = fba.run(reactionsSplit, matrixSplit, lbSplit,
                                           ubSplit, fbaParams)
                # Get joined solution (for output)
                sFlux = array(FbAnalyzer.rejoinFluxes(sFlux, reactionsSplit,
                                                      modelRed.reactionDict))
            else:
                obj_value, sFlux = fba.run(modelRed.getReactionNames(), matrix,
                                           lb, ub, fbaParams)

            if len(sFlux) == 0:
                lbFull, ubFull = list(map(array, model.getBounds()))
                return [], sFlux, lbFull, ubFull, dimReduced

        # Use negative threshold (objective value >= threshold is equivalent to
        # -objective value <= -threshold, and objective already has coefficient
        # -1 due to maximization)
        threshold = maxmin_factor * obj_value * threshp
        print("obj func. opt:", obj_value)
        print("Threshold: ", threshold)
        try:
            eqs, ineqs = ParamParser.linConstraintsToVectors(
                fbaParams.linConstraints, modelRed.reactionDict)
        except ValueError as e:
            # If any linear constraint is contradictory, report error
            print("Optimization not possible due to contradictory constraints:")
            print("  "+e)
            exit()

        minmax = self.run(objective, matrix, lb, ub, threshold, eqs, ineqs)

        # Add removed reactions with min = max = 0. to minmax and solution
        if len(modelRed) != len(model):
            minmaxFull, sFluxFull = [], []
            reactionsRed = set(modelRed.getReactionNames())

            for rea in model:
                if rea.name in reactionsRed:
                    # Take values from minmax and sFlux
                    index = modelRed.reactionDict[rea.name]
                    minmaxFull.append(minmax[index])
                    sFluxFull.append(sFlux[index])
                else:
                    # Dead reaction => min = max = flux = 0.
                    minmaxFull.append((0., 0.))
                    sFluxFull.append(0.)

            lbFull, ubFull = list(map(array, model.getBounds()))
            return minmaxFull, array(sFluxFull), lbFull, ubFull, dimReduced
        else:
            return minmax, sFlux, lb, ub, dimReduced

    @staticmethod
    def writeSolutionToFile(filename, reactions, solution, lb, ub, minmax):
        with open(filename, 'w') as f:
            if filename.split(".")[-1] == "csv":
                FvAnalyzer.writeSolutionToCSVHandle(f, reactions, solution, lb, ub,
                                                    minmax)
            else:
                FvAnalyzer.writeSolutionToFileHandle(f, reactions, solution, lb, ub,
                                                     minmax)

    @staticmethod
    def writeSolutionToCSVHandle(f, reactions, solution, lb, ub, minmax):
        """ write to the CSV file given by file handle f (must be open for writing)
        """
        if len(solution) == 0:
            return  # Nothing to do

        csvwriter = csv.writer(
            f, delimiter=";", quotechar='"', quoting=csv.QUOTE_MINIMAL)
        # Write Header to CSV file
        csvwriter.writerow(["NAME", "MIN_FLUX", "FBA_FLUX",
                            "MAX_FLUX", "DIFF", "LB", "UB"])
        # Construct output table and get widths of table columns (length of
        # longest occurring string)
        for rea in sorted(reactions):
            index = reactions[rea]
            mini, maxi = minmax[index]
            diff = maxi-mini
            flux = solution[index]
            lbVec = lb[index]
            ubVec = ub[index]

            csvwriter.writerow([rea, mini, flux, maxi, diff, lbVec, ubVec])

    @staticmethod
    def writeSolutionToFileHandle(f, reactions, solution, lb, ub, minmax):
        if len(solution) == 0:
            return  # Nothing to do

        # Construct output table and get widths of table columns (length of
        # longest occurring string)
        maxlenName, maxlenMin, maxlenFlux, maxlenMax, maxlenDiff = 0, 0, 0, 0, 0
        maxlenLb, maxlenUb = 0, 0
        #  dict {name : (minFlux, fbaFlux, maxFlux, diff, lb, ub)} - as strings
        outputTable = {}
        for rea in reactions:
            index = reactions[rea]
            mini, maxi = minmax[index]
            # Pad non-negative numbers with space
            outputTable[rea] = tuple(map(padNumber, list(map(repr, (solution[index],
                                                                    lb[index], ub[index], mini, maxi, maxi-mini)))))
            lenName, lenFlux, lenLb, lenUb, lenMin, lenMax, lenDiff = \
                list(map(len, (rea,)+outputTable[rea]))

            if lenName > maxlenName:
                maxlenName = lenName
            if lenFlux > maxlenFlux:
                maxlenFlux = lenFlux
            if lenLb > maxlenLb:
                maxlenLb = lenLb
            if lenUb > maxlenUb:
                maxlenUb = lenUb
            if lenMin > maxlenMin:
                maxlenMin = lenMin
            if lenMax > maxlenMax:
                maxlenMax = lenMax
            if lenDiff > maxlenDiff:
                maxlenDiff = lenDiff

        # Write table head
        f.write("NAME".ljust(maxlenName)+"   "+"MIN_FLUX".rjust(maxlenMin)+" " +
                "FBA_FLUX".rjust(maxlenFlux)+" " +
                "MAX_FLUX".rjust(maxlenMax)+" "
                + "DIFF".rjust(maxlenDiff)+" "+"LB".rjust(maxlenLb)+" " +
                "UB".rjust(maxlenUb)+"\n")

        # Write result for the flux through every reaction
        for rea in sorted(reactions):
            flux, lb, ub, mini, maxi, diff = outputTable[rea]
            f.write(rea.ljust(maxlenName) + " : " + mini.rjust(maxlenMin) + " "
                    + flux.rjust(maxlenFlux) + " " +
                    maxi.rjust(maxlenMax) + " "
                    + diff.rjust(maxlenDiff) + " " + lb.rjust(maxlenLb) + " " +
                    ub.rjust(maxlenUb) + "\n")

    @staticmethod
    def parseSolutionFile(filename):
        """ read the given FVA solution file identified by filename
            -- wrapper for parseSolutionFileByHandle()
        """
        with open(filename) as f:
            if filename.split(".")[-1] == "csv":
                return FvAnalyzer.parseCSVFileByHandle(f)
            else:
                return FvAnalyzer.parseSolutionFileByHandle(f)

    @staticmethod
    def parseCSVFileByHandle(f):
        """ read the FVA solution file given as a file object (must be open)

        Returns:
        (fbaSolution, minmax) tuple with

        fbaSolution -- MetabolicFlux object with flux distribution, and lb, ub
        minmax      -- dict { reaction : flux minimum, flux maximum }
        """
        csvreader = csv.reader(f, delimiter=";", quotechar='"')
        reactions = set()
        reactionVec, fluxVec = [], []
        boundsDict = {}
        minmax = {}

        line_no = 0
        for line in csvreader:
            line_no += 1
            if line[0] == "NAME":
                continue

            rea = line[0]
            if rea in reactions:
                raise SyntaxError("Syntax error in line %u: Duplicate "
                                  "reaction." % line_no)
            reactions.add(rea)
            reactionVec.append(rea)

            try:
                values = list(map(float, line[1:]))
            except ValueError:
                raise SyntaxError("Syntax error in line %u: Invalid "
                                  "floating point value." % line_no)
            try:
                fluxVec.append(values[1])
                minmax[rea] = values[0], values[2]
                boundsDict[rea] = values[4], values[5]
            except IndexError:
                raise SyntaxError("Syntax error in line %u:\nLine must "
                                  "contain exactly six values (min_flux, "
                                  "fba_flux, max_flux, diff, lb, ub)." %
                                  line_no)
        return MetabolicFlux(reactionVec, fluxVec, boundsDict), minmax

    @staticmethod
    def parseSolutionFileByHandle(f):
        """ read the FVA solution file given as a file object (must be open)

        Returns:
        (fbaSolution, minmax) tuple with

        fbaSolution -- MetabolicFlux object with flux distribution, and lb, ub
        minmax      -- dict { reaction : flux minimum, flux maximum }
        """
        reactions = set()
        reactionVec, fluxVec = [], []
        boundsDict = {}
        minmax = {}

        line_no = 0
        for line in f:
            line_no += 1
            if (line.lstrip().upper().startswith("NAME") or line == "" or
                    line.isspace()):
                continue

            try:
                rea, values_str = list(map(str.rstrip, line.split(":")))
            except ValueError:
                raise SyntaxError("Syntax error in line %u:\nLine must "
                                  "contain exactly one colon (':')." %
                                  line_no)
            if rea in reactions:
                raise SyntaxError("Syntax error in line %u: Duplicate "
                                  "reaction." % line_no)
            reactions.add(rea)
            reactionVec.append(rea)

            try:
                values = list(map(float, values_str.split(None, 6)[:6]))
            except ValueError:
                raise SyntaxError("Syntax error in line %u: Invalid "
                                  "floating point value." % line_no)
            try:
                fluxVec.append(values[1])
                minmax[rea] = values[0], values[2]
                boundsDict[rea] = values[4], values[5]
            except IndexError:
                raise SyntaxError("Syntax error in line %u:\nLine must "
                                  "contain exactly six values (min_flux, "
                                  "fba_flux, max_flux, diff, lb, ub)." %
                                  line_no)
        return MetabolicFlux(reactionVec, fluxVec, boundsDict), minmax

    @staticmethod
    def getBlockedReactions(fvaMinmax):
        """ get reactions whose flux is restricted to zero based on FVA solution

        Keyword arguments:
        fvaMinmax  -- dict { reaction : flux minimum, flux maximum }

        Returns list of reactions (in arbitrary order)
        """
        blockedReactions = []
        for name in fvaMinmax:
            mini, maxi = fvaMinmax[name]
            if mini == maxi == 0.:
                blockedReactions.append(name)
        return blockedReactions


def main():
    # 1. Parse command line

    usage = "Usage: %prog [options]"
    version = "Flux variability analysis\n" + COPYRIGHT_VERSION_STRING
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-r", "--reactions", dest="reactionFile",
                      help="perform Flux Variability Analysis on the network "
                           "given by the reaction FILE", metavar="FILE")
    parser.add_option("-p", "--parameters", dest="paramFile",
                      help="use the given scenario FILE for Flux Variability "
                           "Analysis", metavar="FILE")
    parser.add_option("-o", "--output", dest="outputFile",
                      help="write output of Flux Variability Analysis to FILE",
                      metavar="FILE")
    parser.add_option("-t", "--tolerance", dest="tolerance", type="float",
                      help="tolerance for objective function (0 < VALUE "
                      "< 1; default: .95)", metavar="VALUE")
    parser.add_option("-s", "--solution-file", dest="solutionFile", help="read "
                      "FBA solution from FILE (optional)", metavar="FILE")
    parser.add_option("-l", "--use-full-matrix", action="store_true",
                      dest="useFullMatrix", help="use full matrix (disable "
                      "removal of dead ends and nonfunctional reactions)")
    parser.set_defaults(tolerance=.95, useFullMatrix=False)

    options, _ = parser.parse_args()
    parser.check_required("-r")
    parser.check_required("-p")
    parser.check_required("-o")

    if options.tolerance <= 0. or options.tolerance > 1.:
        print("Error: Tolerance must be in interval (0, 1]")
        exit()

    # 2. Parse reaction file

    rparser = ReactionParser()
    model = MetabolicModel()
    try:
        model.addReactionsFromFile(options.reactionFile, rparser)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in reaction file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()

    # 3. Parse scenario file

    model_messages = []
    pparser = ParamParser()
    try:
        # Parse file, get maxmin, name of objective function, and solver name
        maxmin, objStr, solver, numIter, lb, ub = pparser.parse(
            options.paramFile)
        fbaParams = FbaParam(solver, maxmin, objStr, numIter)
        fbaParams.setLinConstraints(pparser.lin_constraints)
        # Set flux bounds in model
        model.setFiniteBounds(lb, ub, True, model_messages)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.paramFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in scenario file %s:" %
               os.path.basename(options.paramFile))
        print(strerror)
        exit()
    except ValueError as strerror:
        print(strerror)
        exit()

    # Show warning and info messages of parsers
    msgs = (rparser.getMessages() + pparser.getMessages() +
            [x[1] for x in model_messages])
    if msgs:
        print('\n'+'\n'.join(msgs))

    # 4. Launch flux variability analysis

    print("Launching flux variability analysis.")
    print("tolerance: %g" % options.tolerance)
    if solver == "":
        solver = "default"
    fva = FvAnalyzer(solver)
    minmax, sFlux, lb, ub, ndim = fva.runOnModel(model, fbaParams,
                                                 options.tolerance, options.solutionFile,
                                                 rmDeadEnds=not options.useFullMatrix)

    print ("Info: The reduced network has %u reactions and %u metabolites." %
           ndim[1::-1])
    if not minmax:
        print("Model is infeasible or unbounded. Nothing to do.")
        exit()

    # 5. Write output to file

    try:
        fva.writeSolutionToFile(options.outputFile, model.reactionDict, sFlux,
                                lb, ub, minmax)
    except IOError as strerror:
        print ("Unable to write to file %s:" %
               os.path.basename(options.outputFile))
        print(strerror)
        exit()


if __name__ == "__main__":
    main()




'''
This module evaluates the objective function by setting each
metabolite as export objective function.
As result this file prints all metabolites that could not be
produced by the network.

This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
'''
from __future__ import print_function

from builtins import object
import optparse
import os

from metano.metabolicmodel import MetabolicModel
from metano.reactionparser import ReactionParser
from metano.paramparser import ParamParser
from metano.fba import FbAnalyzer
from metano.defines import FbaParam, COPYRIGHT_VERSION_STRING


class OptionParser(optparse.OptionParser):
    """ Extension of OptionParser with function for checking required arguments
    """

    def check_required(self, opt):
        """ check if required options are given """
        option = self.get_option(opt)

        # Assumes the option's 'default' is set to None!
        if getattr(self.values, option.dest) is None:
            self.error("%s option not supplied" % option)


class EvalObjective(object):
    """ Evaluates objective function by testing each metabolite """
    def __init__(self, model, scenario):
        self.model = model
        self.scenario = scenario
        self.run_count = 0
        self.result = []

        objfunindex = model.reactionDict[model.biomass_name]
        metabolites = model.reactions[objfunindex].metabolites

        self.objfun_dict = {}
        for met in metabolites:
            self.objfun_dict[met.name] = met.coef

    def run(self):
        """ start evaluation process and save results """
        print ("Start evaluation of %s metabolites from "
               "the objective function." % len(self.objfun_dict))
        for metabolite in self.objfun_dict:
            pparser = ParamParser()
            maxmin, objstr, solver, numiter, lb , ub = pparser.parse(self.scenario)

            scenariomodel = MetabolicModel()
            scenariomodel.addReactions(self.model.reactions)
            scenariomodel.addReaction(("temp_obj_fun", ((1, metabolite),), "-->", ()))
            #objfunindex = scenariomodel.reactionDict["temp_obj_fun"]

            fbaparams = FbaParam(solver, True, "temp_obj_fun", numiter)
            fbaparams.setLinConstraints(pparser.lin_constraints)

            scenariomodel.setFiniteBounds(lb, ub, True)
            fba = FbAnalyzer(verbose=False)
            obj_value, solution, _ = fba.runOnModel(scenariomodel, fbaparams, rmDeadEnds=True)
            if not solution:
                self.result.append(metabolite)


    def print_results(self):
        """ print results from run """
        if not self.result:
            print("All metabolites from the objective functions are produced by the network!")
            return
        print("-----------------------------------------------------------------------")
        print("RESULT: the following metabolites could not be produced in the network:")
        for i in sorted(self.result):
            print("\t" + i)
        print("-----------------------------------------------------------------------")


def main():
    """ main function, parse command line and pass options to class """
    # 1. Parse command line

    usage = "Usage: %prog [options]"
    version = "multiple Flux balance analysis\n" + COPYRIGHT_VERSION_STRING
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-r", "--reactions", dest="reactionFile",
                      help="perform multiple Flux Balance Analysis on the network given "
                           "by the reaction FILE", metavar="FILE")
    parser.add_option("-p", "--parameters", dest="paramFile",
                      help="basic scenario file", metavar="FILE")

    options, _ = parser.parse_args()
    parser.check_required("-r")
    parser.check_required("-p")
    #parser.check_required("-f")

    # 1. parse reaction file
    rparser = ReactionParser()
    model = MetabolicModel()
    try:
        model.addReactionsFromFile(options.reactionFile, rparser)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in reaction file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()

    # 2. evaluate objective function
    evalobj = EvalObjective(model, options.paramFile)
    evalobj.run()
    evalobj.print_results()



if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
This module defines a parser for SBML files. It is also a script for converting
SBML files to metano files. This parser ignores units. The most recent version
supported is SBML Level 2 Version 4, with the Flux Balance Constraints
extension: http://sbml.org/Documents/Specifications/SBML_Level_3/Packages/Flux_Balance_Constraints_(flux).


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

Alexander Krause contributed to this file.

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function
from __future__ import absolute_import

from builtins import zip
from builtins import object
from metano.defines import (Verbosity, ReaFileStruc, makeCoefNameTerm, isZeroVector,
                     COPYRIGHT_VERSION_STRING)
from metano.sbmlstructure import SbmlDef, getFirstElementChild
from metano.fba import OptionParser
from metano.metabolicmodel import Reactant, Reaction, MetabolicModel
from metano.metabolicflux import MetabolicFlux
import xml.dom.minidom as dom
from xml.parsers.expat import ExpatError
from numpy import inf
import os, re


class SbmlParser(object):
    """ Class for parsing SBML files

    Member variables:

    modelName          -- name of SBML model
    listOfReactions    -- list of Reaction objects (class Reaction defined in
                          metabolicmodel.py)
    listOfFluxValues   -- list of fluxes corresponding to the reactions in
                          listOfReactions (left empty, if no fluxes are found)
    listOfObjCoef      -- list of OBJECTIVE_COEFFICIENT values for each reaction
    setOfReactions     -- IDs of reactions already read
    dictOfCompartments -- dict { ID : (name, outside) } for each compartment
    dictOfSpecies      -- stores tuple (name, compartment, boundaryCondition)
                          for each species ID
    dictOfReactions    -- dict { ID : name } for each reaction
    infLevel           -- threshold above which a bound is considered infinite
                          (default: 1000)
    messages           -- error, warning, info, and debug messages generated
                          during parsing
    """

    _decodeBool = {"false": 0, "False": 0, "0": 0, "true": 1, "True": 1, "1": 1}


    def __init__(self, infLevel=1000.):
        """ initialize class SbmlParser
        """
        self.infLevel = infLevel

    def clear(self):
        """ clear all internal variables, warnings and info messages
        """
        self.messages = []
        self.modelName = ""
        self.listOfReactions = []
        self.setOfReactions = set()
        self.listOfFluxValues = []
        self.listOfObjCoef = []
        self.dictOfCompartments = {}
        self.dictOfSpecies = {}
        self.dictOfReactions = {}
        self._tmpReaction = CacheReaction()

    def getMetabolicModelComponents(self):
        """ return structures needed to build a MetabolicModel
        """
        # Return only non-boundary species, with 'isBoundary' field removed
        metaboliteLabels = dict((ID, (name, compartment)) for
                                (ID, (name, compartment, isBoundary)) in
                                list(self.dictOfSpecies.items()) if not isBoundary)

        return self.listOfReactions, metaboliteLabels, self.dictOfReactions

    def getFluxValues(self, reaNameDict=None):
        """ return the flux distribution read from the SBML file as dict
            { reaction ID/name : flux value }

        Keyword arguments:

        reaNameDict -- dict {reaction ID : name } (optional; if present,
                                                   names must be unique)
        """
        if reaNameDict:
            return dict(list(zip([reaNameDict[x.name] for x in self.listOfReactions],
                            self.listOfFluxValues)))
        else:
            return dict(list(zip([x.name for x in self.listOfReactions],
                            self.listOfFluxValues)))

    def getObjCoefficients(self, reaNameDict=None):
        """ return the coefficient vector of the linear objective function read
            from the SBML file as dict { reaction ID/name : coefficient }

        Keyword arguments:

        reaNameDict -- dict {reaction ID : name } (optional; if present,
                                                   names must be unique)
        """
        if reaNameDict:
            return dict(list(zip([reaNameDict[x.name] for x in self.listOfReactions],
                            self.listOfObjCoef)))
        else:
            return dict(list(zip([x.name for x in self.listOfReactions],
                            self.listOfObjCoef)))

    def getMessages(self, level=Verbosity.INFO):
        """ return a list of all messages at or above the given level of
            severity

        Levels are defined in defines.py.
        """
        return [x[1] for x in self.messages if x[0] <= level]


    # Shortcuts for generating common warnings, error and debug messages

    def _errorEmptyCollection(self, collectionName):
        self.messages.append((Verbosity.ERROR,
                              "Error: %s is empty." % collectionName))
    def _errorNoId(self, nodeType):
        self.messages.append((Verbosity.ERROR,
            "Error: Found %s with no ID or name." % nodeType))
    def _errorDuplicateId(self, ID, collectionName):
        self.messages.append((Verbosity.ERROR,
            "Error: Found duplicate ID (%s) in %s." % (ID, collectionName)))
    def _errorIllegalFloat(self, node, upperType, upperId):
        self.messages.append((Verbosity.ERROR, "Error: Illegal floating point "
                              "value for %s in %s '%s'." % (node, upperType,
                                                            upperId)))
    def _errorMissingValue(self, node, upperType, upperId):
        self.messages.append((Verbosity.ERROR, "Error in %s '%s': Parameter "
            "%s has no %s." % (upperType, upperId, node, SbmlDef.VALUE)))

    def _warnDuplicateNode(self, node, upperType="", upperId=""):
        msg = "Warning: Duplicate %s found" % node
        if upperId:
            msg += " in %s '%s'" % (upperType, upperId)
        msg += ". Only the first is used."
        self.messages.append((Verbosity.WARNING, msg))
    def _warnIllegalValueUsingDefault(self, nodeType, ID, attribute, default):
        self.messages.append((Verbosity.WARNING,
            "Warning: Unknown value encountered for attribute \"%s\" in %s "
            "'%s'. Set to default: %s." % (attribute, nodeType, ID, default)))

    def _debugAttributeValue(self, attribute, value):
        self.messages.append((Verbosity.DEBUG,
                              "Debug: %s = %s." % (attribute, value)))
    def _debugUsingIdAsName(self, nodeType, ID, nameStr):
        self.messages.append((Verbosity.DEBUG, "Debug: %s '%s' is missing "
                              "attribute \"%s\". Using ID instead." %
                              (nodeType.capitalize(), ID, nameStr)))
    def _debugUsingNameAsId(self, nodeType, name):
        self.messages.append((Verbosity.DEBUG, "Debug: %s named '%s' has no ID."
                              " Using name instead." %
                              (nodeType.capitalize(), name)))
    def _debugIgnoringNode(self, node, upperId="", upperType=""):
        msg = "Debug: Ignoring SBML node \"%s\"" % node
        if upperId:
            if upperType:
                msg += " under %s '%s'" % (upperType, upperId)
            else:
                msg += " in %s" % upperId
        msg += "."
        self.messages.append((Verbosity.DEBUG, msg))


    # Main methods

    def readFile(self, filename, boundaryRegex=""):
        """ read the given SBML file into member structures

        Keyword arguments:

        filename      -- name of the SBML file to be parsed
        boundaryRegex -- if given, species with ID matching this regular
                         expression are treated as additional boundary species

        Returns True if successful, else False
        """
        self.clear()
        isModelFound = False
        fc = dom.parse(filename)

        # Read model node first
        # There are ELEMENT_NODES with subtrees and TEXT_NODES - only the
        # ELEMENT_NODES are processed
        for node in getFirstElementChild(fc).childNodes:
            if node.nodeName.lower() == SbmlDef.MODEL.lower():
                isModelFound = True

                # Read model name (optional)
                if node.getAttribute(SbmlDef.NAME) != "":
                    self.modelName = node.getAttribute(SbmlDef.NAME)
                    msg = ("Info: The metabolic model has the name '%s'."
                           % self.modelName)
                    self.messages.append((Verbosity.INFO, msg))
                else:
                    msg = "Info: The metabolic model has no name."
                    self.messages.append((Verbosity.INFO, msg))

                # Read the model elements
                success = self._readModelNodes(node)
                break  # an SBML file may contain only one model

        if not isModelFound:
            success = False
            msg = "Error: Unable to read model."
            self.messages.append((Verbosity.ERROR, msg))

        if success:
            nBoundary = len([isBoundary for (_, _, isBoundary) in
                             list(self.dictOfSpecies.values()) if isBoundary])
            msg = ("Info: The SBML model has %d reactions and %d species" %
                   (len(self.listOfReactions), len(self.dictOfSpecies)))
            if nBoundary:
                msg += " (including %d boundary species)" % nBoundary
            msg +="."
            self.messages.append((Verbosity.INFO, msg))

            if boundaryRegex:
                try:
                    pattern = re.compile(boundaryRegex)
                except re.error as strerror:
                    msg = ("Warning: Error in regular expression for boundary "
                           "metabolites: %s. Unable to evaluate." %
                           strerror)
                    self.messages.append((Verbosity.WARNING, msg))
                    boundaryRegex = ""
                    pattern = None

                if pattern:
                    # Delete all metabolites matching the given reg-ex
                    for reaction in self.listOfReactions:
                        for reactant in reaction:
                            metName = reactant.name
                            if pattern.search(metName):
                                reaction.removeReactant(metName)
                                msg = ("Debug: Deleted boundary metabolite '%s'"
                                       " from reaction '%s'." % (metName,
                                                                 reaction.name))
                                self.messages.append((Verbosity.DEBUG, msg))

        return success


    def fixNames(self):
        """ modify reaction, species, and compartment IDs and names by replacing
            all characters not allowed in metano reaction files with underscores
            and converting all strings to type str
        """
        illegC = re.compile("\\s|%s|%s" % (ReaFileStruc.delimiter,
                                           ReaFileStruc.commentSign))
        # Reaction IDs and species IDs in reactions
        for reaction in self.listOfReactions:
            reaction.name = re.sub(illegC, "_", reaction.name).encode('ascii',
                                                                      'replace')
            for reactant in reaction:
                reactant.name = re.sub(illegC, "_", reactant.name).encode(
                                                   'ascii', 'replace')

        # Reaction names
        strDictOfReactions = {}
        for rea in self.dictOfReactions:
            strDictOfReactions[re.sub(illegC, "_", rea).encode('ascii',
                'replace')] = re.sub(illegC, "_",
                self.dictOfReactions[rea]).encode('ascii', 'replace')

        self.dictOfReactions = strDictOfReactions

        strDictOfSpecies = {}
        for met in self.dictOfSpecies:
            # Species names and compartment IDs in species
            name, compartment, isBoundary = self.dictOfSpecies[met]
            strDictOfSpecies[re.sub(illegC, "_", met).encode('ascii',
                'replace')] = (re.sub(illegC, "_", name).encode('ascii',
                'replace'), re.sub(illegC, "_", compartment).encode('ascii',
                'replace'), isBoundary)

        self.dictOfSpecies = strDictOfSpecies

        strDictOfCompartments = {}
        for cId in self.dictOfCompartments:
            # Name of compartment and ID of outside compartment
            name, outside = self.dictOfCompartments[cId]
            strDictOfCompartments[re.sub(illegC, "_", cId).encode('ascii',
                'replace')] = (re.sub(illegC, "_", name).encode('ascii',
                'replace'), re.sub(illegC, "_", outside).encode('ascii',
                                                                'replace'))

        self.dictOfCompartments = strDictOfCompartments


    def buildModel(self, usenames=False):
        """ return a MetabolicModel constructed from the data read from the SBML
            file

        Keyword arguments:

        usenames -- if True, write reaction and species names as labels to the
                    model
        """
        reactions, metaboliteLabels, reactionLabels = \
            self.getMetabolicModelComponents()

        model = MetabolicModel()
        model.addReactions(reactions)

        # Set labels built from name and compartment if names are to be used
        # instead of IDs
        if usenames:
            metLabelDict = dict((k, "%s[%s]" % x)
                                for (k, x) in list(metaboliteLabels.items()))
            model.setMetaboliteLabels(metLabelDict)
            model.setReactionLabels(reactionLabels)
            # Make sure that all reaction and metabolite labels are unique
            model.makeLabelsUnique()

        return model


    # Auxiliary methods called by readFile() to parse subtrees

    def _readModelNodes(self, modelNode):
        """ read the subtree under the model node

        Returns True if successful, else False
        """
        isListOfCompartmentsFound = False
        isListOfCompartmentsRead = False
        isListOfSpeciesFound = False
        isListOfSpeciesRead = False
        isListOfReactionsFound = False
        isListOfReactionsRead = False

        # Notes:
        # 1) If the model contains more than one listOfCompartments,
        #    listOfSpecies, or listOfReactions, only the first occurrence of
        #    each is processed.
        # 2) Order is important in SBML: listOfSpecies cannot occur before
        #    listOfCompartments, and listOfReactions cannot occur before
        #    listOfSpecies

        for node in modelNode.childNodes:
            nodeName = node.nodeName.lower()

            if nodeName == SbmlDef.LISTOFCOMPARTMENTS.lower():

                if isListOfCompartmentsFound:
                    self._warnDuplicateNode(SbmlDef.LISTOFCOMPARTMENTS)
                else:
                    isListOfCompartmentsFound = True
                    isListOfCompartmentsRead = \
                        self._readListOfCompartments(node)

            elif nodeName == SbmlDef.LISTOFSPECIES.lower():

                if isListOfSpeciesFound:
                    self._warnDuplicateNode(SbmlDef.LISTOFSPECIES)

                elif not isListOfCompartmentsRead:
                    msg = ("Error: Encountered %s, but need %s first. "
                           "Unable to process." % (SbmlDef.LISTOFSPECIES,
                                                   SbmlDef.LISTOFCOMPARTMENTS))
                    self.messages.append((Verbosity.ERROR, msg))
                    isListOfSpeciesFound = True
                else:
                    isListOfSpeciesFound = True
                    isListOfSpeciesRead = self._readListOfSpecies(node)

            elif nodeName == SbmlDef.LISTOFREACTIONS.lower():

                if isListOfReactionsFound:
                    self._warnDuplicateNode(SbmlDef.LISTOFREACTIONS)

                elif not isListOfSpeciesRead:
                    msg = ("Error: Encountered %s, but need %s first. "
                           "Unable to process." % (SbmlDef.LISTOFREACTIONS,
                                                   SbmlDef.LISTOFSPECIES))
                    self.messages.append((Verbosity.ERROR, msg))
                    isListOfReactionsFound = True
                else:
                    isListOfReactionsFound = True
                    isListOfReactionsRead = self._readListOfReactions(node)

            elif node.nodeType == dom.Node.ELEMENT_NODE and nodeName != "":
                # Anything that is not a listOfCompartments, listOfSpecies, or
                # listOfReactions is ignored
                self._debugIgnoringNode(node.nodeName)

        if (isListOfSpeciesRead and isListOfReactionsRead and
            isListOfCompartmentsRead):
            return True

        if not isListOfCompartmentsFound:
            self.messages.append((Verbosity.ERROR,
                "Error: %s not found." % SbmlDef.LISTOFCOMPARTMENTS))
        elif not isListOfSpeciesFound:
            self.messages.append((Verbosity.ERROR,
                "Error: %s not found." % SbmlDef.LISTOFSPECIES))
        elif not isListOfReactionsFound:
            self.messages.append((Verbosity.ERROR,
                "Error: %s not found." % SbmlDef.LISTOFREACTIONS))

        return False


    def _readListOfCompartments(self, compartmentsNode):
        """ read the subtree under the listOfCompartments node

        Returns True if successful, else False
        """
        success = True
        for node in compartmentsNode.childNodes:

            if node.nodeName.lower() == SbmlDef.COMPARTMENT.lower():
                compartmentOutside = ""
                compartmentName = ""

                if node.getAttribute(SbmlDef.NAME) != "":
                    compartmentName = node.getAttribute(SbmlDef.NAME)

                if node.getAttribute(SbmlDef.ID) != "":
                    compartmentId = node.getAttribute(SbmlDef.ID)
                elif compartmentName != "":
                    compartmentId = compartmentName
                    self._debugUsingNameAsId("compartment", compartmentName)
                else:
                    self._errorNoId("compartment")
                    success = False
                    continue

                if node.getAttribute(SbmlDef.COMPARTMENTOUTSIDE) != "":
                    compartmentOutside = \
                        node.getAttribute(SbmlDef.COMPARTMENTOUTSIDE)

                # ID must be unique
                if compartmentId in self.dictOfCompartments:
                    self._errorDuplicateId(compartmentId,
                                           SbmlDef.LISTOFCOMPARTMENTS)
                    success = False
                    continue

                self.dictOfCompartments[compartmentId] = (compartmentName,
                                                          compartmentOutside)

                msg = "Info: Found compartment '%s'" % compartmentId
                if compartmentName:
                    msg += " (%s)" % compartmentName
                if compartmentOutside:
                    msg += " with %s='%s'" % (SbmlDef.COMPARTMENTOUTSIDE,
                                              compartmentOutside)
                msg += "."
                self.messages.append((Verbosity.INFO, msg))

            elif node.nodeType == dom.Node.ELEMENT_NODE and node.nodeName != "":
                # Anything that is not a compartment is ignored
                self._debugIgnoringNode(node.nodeName,
                                        SbmlDef.LISTOFCOMPARTMENTS)

        if not self.dictOfCompartments:
            self._errorEmptyCollection(SbmlDef.LISTOFCOMPARTMENTS)
            success = False

        return success


    def _readListOfSpecies(self, listOfSpeciesNode):
        """ read the subtree under the listOfSpecies node

        Returns True if successful, else False
        """
        success = True
        for node in listOfSpeciesNode.childNodes:

            if node.nodeName.lower() == SbmlDef.SPECIES.lower():
                speciesName = ""
                speciesCompartment = ""
                speciesBoundaryCondition = 0

                if node.getAttribute(SbmlDef.NAME) != "":
                    speciesName = node.getAttribute(SbmlDef.NAME)

                # Species must always have ID/name and compartment
                if node.getAttribute(SbmlDef.ID) != "":
                    speciesId = node.getAttribute(SbmlDef.ID)
                elif speciesName != "":
                    speciesId = speciesName
                    self._debugUsingNameAsId("species", speciesName)
                else:
                    self._errorNoId("species")
                    success = False
                    continue

                # Species ID must be unique
                if speciesId in self.dictOfSpecies:
                    self._errorDuplicateId(speciesId, SbmlDef.LISTOFSPECIES)
                    success = False
                    continue

                if node.getAttribute(SbmlDef.COMPARTMENT) != "":
                    speciesCompartment = node.getAttribute(SbmlDef.COMPARTMENT)
                else:
                    msg = ("Error: Species '%s' is missing attribute \"%s\"."
                           % (speciesId, SbmlDef.COMPARTMENT))
                    self.messages.append((Verbosity.ERROR, msg))
                    success = False
                    continue

                msg = "Debug: Found species '%s'" % speciesId
                if speciesName:
                    msg += " (%s)" % speciesName
                msg += " in %s '%s'." % (SbmlDef.COMPARTMENT,
                                         speciesCompartment)
                self.messages.append((Verbosity.DEBUG, msg))

                if not speciesName:
                    speciesName = speciesId
                    self._debugUsingIdAsName("species", speciesId,
                                             SbmlDef.NAME)

                # Process Boolean attribute 'boundaryCondition'
                bcStr = node.getAttribute(SbmlDef.BOUNDARYCONDITION)
                if bcStr != "":
                    if bcStr in self._decodeBool:
                        speciesBoundaryCondition = self._decodeBool[bcStr]
                        self._debugAttributeValue(SbmlDef.BOUNDARYCONDITION,
                                                  speciesBoundaryCondition)
                    else:
                        self._warnIllegalValueUsingDefault("species", speciesId,
                            SbmlDef.BOUNDARYCONDITION, False)

                # Is compartment known?
                if speciesCompartment not in self.dictOfCompartments:
                    msg = ("Error: Undefined %s (%s) found for species '%s.'" %
                           (SbmlDef.COMPARTMENT, speciesCompartment, speciesId))
                    self.messages.append((Verbosity.ERROR, msg))
                    success = False
                    continue

                self.dictOfSpecies[speciesId] = (speciesName,
                               speciesCompartment, speciesBoundaryCondition)

            elif node.nodeType == dom.Node.ELEMENT_NODE and node.nodeName != "":
                # Anything that is not a species is ignored
                self._debugIgnoringNode(node.nodeName, SbmlDef.LISTOFSPECIES)

        if not self.dictOfSpecies:
            self._errorEmptyCollection(SbmlDef.LISTOFSPECIES)
            success = False

        return success


    def _readListOfReactions(self, listOfReactionsNode):
        """ read subtree under the listOfReactions node

        Returns True if successful, else False
        """
        success = True
        listIsEmpty = True
        for node in listOfReactionsNode.childNodes:
            if listIsEmpty:
                listIsEmpty = False
            if node.nodeName.lower() == SbmlDef.REACTION.lower():
                if not self._readReaction(node):
                    success = False

            elif node.nodeType == dom.Node.ELEMENT_NODE and node.nodeName != "":
                # Anything that is not a reaction is ignored
                self._debugIgnoringNode(node.nodeName, SbmlDef.LISTOFREACTIONS)

        if listIsEmpty:
            self._errorEmptyCollection(SbmlDef.LISTOFREACTIONS)
            success = False

        return success


    def _readReaction(self, reactionNode):
        """ process reaction node

        Returns True if successful, else False
        """
        success = True
        reactionId = ""
        reactionName = ""
        isReversible = 1
        isListOfReactantsFound = False
        isListOfProductsFound = False
        isKineticLawFound = False

        # Read reaction ID or generate if not present
        if reactionNode.getAttribute(SbmlDef.ID) != "":
            reactionId = reactionNode.getAttribute(SbmlDef.ID)
            msg = "Debug: Reaction ID is '%s'." % reactionId
            self.messages.append((Verbosity.DEBUG, msg))
        else:
            reactionId = "rea%03d" % (len(self.listOfReactions)+1)
            msg = ("Warning: Found reaction without attribute \"%s\". Generated"
                   " ID: '%s'." % (SbmlDef.ID, reactionId))
            self.messages.append((Verbosity.WARNING, msg))

        # ID must be unique
        if reactionId in self.setOfReactions:
            self._errorDuplicateId(reactionId, SbmlDef.LISTOFREACTIONS)
            success = False
        else:
            self.setOfReactions.add(reactionId)

        # Read reaction name (optional)
        if reactionNode.getAttribute(SbmlDef.NAME) != "":
            reactionName = reactionNode.getAttribute(SbmlDef.NAME)
            msg = "Debug: Reaction name is '%s'." % reactionName
            self.messages.append((Verbosity.DEBUG, msg))
        else:
            reactionName = reactionId
            self._debugUsingIdAsName("reaction", reactionId, SbmlDef.NAME)

        # Process Boolean attribute 'reversible'
        reversibleStr = reactionNode.getAttribute(SbmlDef.REVERSIBLE)
        if reversibleStr != "":
            if reversibleStr in self._decodeBool:
                isReversible = self._decodeBool[reversibleStr]
                self._debugAttributeValue(SbmlDef.REVERSIBLE, isReversible)
            else:
                self._warnIllegalValueUsingDefault("reaction", reactionId,
                                                   SbmlDef.REVERSIBLE, True)
        else:
            msg = ("Debug: Attribute \"%s\" in reaction '%s' is empty or "
                   "missing. Set to default: True." % (SbmlDef.REVERSIBLE,
                                                       reactionId))
            self.messages.append((Verbosity.DEBUG, msg))

        self._tmpReaction.reactionId = reactionId
        self._tmpReaction.reactionName = reactionName
        self._tmpReaction.reversible = isReversible

        # Process child nodes (listOfReactants, listOfProducts, kineticLaw,
        # notes) - data is exchanged via self._tmpReaction
        for node in reactionNode.childNodes:
            nodeName = node.nodeName.lower()

            if nodeName == SbmlDef.LISTOFREACTANTS.lower():

                if isListOfReactantsFound:
                    self._warnDuplicateNode(SbmlDef.LISTOFREACTANTS, "reaction",
                                            reactionId)
                else:
                    isListOfReactantsFound = True
                    if not self._readListOfReactantsProducts(node, "reactant"):
                        success = False

            elif nodeName == SbmlDef.LISTOFPRODUCTS.lower():

                if isListOfProductsFound:
                    self._warnDuplicateNode(SbmlDef.LISTOFPRODUCTS, "reaction",
                                            reactionId)
                else:
                    isListOfProductsFound = True
                    if not self._readListOfReactantsProducts(node, "product"):
                        success = False

            elif nodeName == SbmlDef.KINETICLAW.lower():

                if isKineticLawFound:
                    self._warnDuplicateNode(SbmlDef.KINETICLAW, "reaction",
                                            reactionId)
                else:
                    isKineticLawFound = True
                    if not self._readKineticLaw(node):
                        success = False

            elif nodeName == SbmlDef.NOTES.lower():
                self._tmpReaction.notes = self._extractTextFromSubtree(node)

            elif node.nodeType == dom.Node.ELEMENT_NODE and nodeName != "":
                self._debugIgnoringNode(node.nodeName, reactionId, "reaction")

        isListOfReactantsFound = (isListOfReactantsFound and
                                  self._tmpReaction.reactants)
        isListOfProductsFound = (isListOfProductsFound and
                                 self._tmpReaction.products)

        # Check for completeness
        if not isListOfReactantsFound:
            if not isListOfProductsFound:
                msg = ("Error: Reaction '%s' has neither reactants nor "
                       "products." % reactionId)
                self.messages.append((Verbosity.ERROR, msg))
                success = False
            else:
                msg = ("Debug: Reaction '%s' has no reactants (left-hand side is"
                       " empty)." % reactionId)
                self.messages.append((Verbosity.DEBUG, msg))
        elif not isListOfProductsFound:
            msg = ("Debug: Reaction '%s' has no products (right-hand side is "
                   "empty)." % reactionId)
            self.messages.append((Verbosity.DEBUG, msg))

        if not isKineticLawFound:
            msg = ("Warning: Reaction '%s' has no parameters (bounds)." %
                   reactionId)
            self.messages.append((Verbosity.WARNING, msg))

        # If processing of child nodes was successful, create Reaction object
        if success:
            comment = self._tmpReaction.notes.strip()
            rea = Reaction(reactionId, self._tmpReaction.reactants,
                           products=self._tmpReaction.products, comment=comment,
                           lb=self._tmpReaction.lb, ub=self._tmpReaction.ub,
                           direction=1-self._tmpReaction.reversible)

            self.listOfReactions.append(rea)
            self.listOfFluxValues.append(self._tmpReaction.fluxVal)
            self.listOfObjCoef.append(self._tmpReaction.objCoef)
            self.dictOfReactions[reactionId] = reactionName

        self._tmpReaction.clear()
        return success


    def _readListOfReactantsProducts(self, listNode, typ):
        """ process listOfReactants or listOfProducts node

        typ is either 'reactant' or 'product'.

        Returns True if successful, else False
        """
        success = True
        reactionId = self._tmpReaction.reactionId
        reactantSet = set()

        for node in listNode.childNodes:
            if node.nodeName.lower() == SbmlDef.SPECIESREFERENCE.lower():
                speciesId = ""
                coef = 0.

                # Read relevant attributes (ID and stoichiometric coefficient)
                if not node.getAttribute(SbmlDef.SPECIES) == "":
                    speciesId = node.getAttribute(SbmlDef.SPECIES)
                else:
                    msg = ("Error: %s without \"%s\" attribute in reaction "
                           "'%s'." % (typ.capitalize(), SbmlDef.SPECIES,
                                      reactionId))
                    self.messages.append((Verbosity.ERROR, msg))
                    success = False
                    continue

                # Species ID must refer to an entry in listOfSpecies
                if speciesId not in self.dictOfSpecies:
                    msg = ("Error: %s '%s' encountered in reaction '%s' is not "
                           "present in %s." % (typ.capitalize(), speciesId,
                                             reactionId, SbmlDef.LISTOFSPECIES))
                    self.messages.append((Verbosity.ERROR, msg))
                    success = False
                    continue

                if node.getAttribute(SbmlDef.STOICHIOMETRY) != "":
                    try:
                        coef = float(node.getAttribute(SbmlDef.STOICHIOMETRY))
                    except ValueError:
                        msg = ("Error: %s '%s' in reaction '%s' has illegal "
                               "\"%s\" value (expected a float)." %
                               (typ.capitalize(), speciesId, reactionId,
                                SbmlDef.STOICHIOMETRY))
                        self.messages.append((Verbosity.ERROR, msg))
                        success = False
                    if coef < 0.:
                        msg = ("Error: Negative stoichiometric coefficient "
                               "encountered for %s '%s' in reaction '%s'." %
                               (typ, speciesId, reactionId))
                        self.messages.append((Verbosity.ERROR, msg))
                        success = False
                else:
                    coef = 1.
                    msg = ("Debug: Attribute \"%s\" of %s '%s' in reaction "
                           "'%s' is empty or missing. Set to default: 1." %
                           (SbmlDef.STOICHIOMETRY, typ, speciesId, reactionId))
                    self.messages.append((Verbosity.DEBUG, msg))

                # Skip reactant/product if it is a boundary species
                name, _, isBoundary = self.dictOfSpecies[speciesId]
                if isBoundary:
                    msg = ("Debug: Skipping boundary species '%s' (%s)." %
                           (speciesId, name))
                    self.messages.append((Verbosity.DEBUG, msg))
                else:
                    reactantSet.add(speciesId)
                    if success:
                        if typ == "reactant":
                            self._tmpReaction.addReactant(Reactant(speciesId,
                                                                   coef))
                        elif typ == "product":
                            self._tmpReaction.addProduct(Reactant(speciesId,
                                                                  coef))
                        msg = ("Debug: Added %s '%s' with coefficient %r." %
                               (typ, speciesId, coef))
                        self.messages.append((Verbosity.DEBUG, msg))

            elif node.nodeType == dom.Node.ELEMENT_NODE and node.nodeName != "":
                # Anything that is not a speciesReference is ignored
                if typ == "product":
                    collectionName = SbmlDef.LISTOFPRODUCTS
                else:
                    collectionName = SbmlDef.LISTOFREACTANTS
                self._debugIgnoringNode(node.nodeName, "%s of reaction '%s'" %
                                        collectionName, reactionId)

        return success


    @staticmethod
    def _extractTextFromSubtree(root):
        text = ""
        for node in root.childNodes:
            if node.nodeType == dom.Node.ELEMENT_NODE:
                text += SbmlParser._extractTextFromSubtree(node)
            else:
                try:
                    s = node.data.strip()
                    if s:
                        text += s + '\n'
                except AttributeError:
                    pass
        return text


    def _readKineticLaw(self, kineticLawNode):
        """ read subtree under a reaction's 'kineticLaw' node

        Returns True if successful, else False
        """
        isListOfParametersFound = False
        success = True
        reactionId = self._tmpReaction.reactionId

        # kineticLaw has two possible subnodes, math and listOfParameters.
        # We ignore the math node and look for definitions of lower bound,
        # upper bound, flux value, etc. in the listOfParameters.
        for node in kineticLawNode.childNodes:
            nodeName = node.nodeName.lower()

            if nodeName == SbmlDef.LISTOFPARAMETERS.lower():
                if isListOfParametersFound:
                    # If the reaction contains more than one listOfParameters,
                    # only the first occurrence is processed.
                    self._warnDuplicateNode(SbmlDef.LISTOFPARAMETERS,
                                            "reaction", reactionId)
                else:
                    isListOfParametersFound = True
                    success = self._readListOfParameters(node)

            elif (node.nodeType == dom.Node.ELEMENT_NODE and nodeName != ""
                  and nodeName != SbmlDef.MATH.lower()):
                # Report anything that is not math or listOfParameters
                self._debugIgnoringNode(node.nodeName, "reaction", reactionId)

        return success


    def _readListOfParameters(self, listOfParametersNode):
        """ read subtree under the 'listOfParameters' node of a 'kineticLaw'

        This function does not deal with standard SBML, but rather with a
        specific extension for flux balance models.

        Returns True if successful, else False
        """
        isLowerBoundFound = False
        isLowerBoundRead = False
        isUpperBoundFound = False
        isUpperBoundRead = False
        isFluxValFound = False
        isFluxValRead = False
        isObjCoefFound = False
        isObjCoefRead = False
        reactionId = self._tmpReaction.reactionId

        lbDefString = SbmlDef.P_LOWER_BOUND.lower()
        ubDefString = SbmlDef.P_UPPER_BOUND.lower()
        fluxDefString = SbmlDef.P_FLUX_VALUE.lower()
        objCoefDefString = SbmlDef.P_OBJ_COEF.lower()

        for node in listOfParametersNode.childNodes:
            if node.nodeName.lower() == SbmlDef.PARAMETER.lower():
                nodeIdOrig = node.getAttribute(SbmlDef.ID)
                nodeName = node.getAttribute(SbmlDef.NAME).lower()
                nodeId = nodeIdOrig.lower()

                if nodeId == lbDefString or nodeName == lbDefString:
                    if isLowerBoundFound:
                        self._warnDuplicateNode(SbmlDef.P_LOWER_BOUND,
                                                "reaction", reactionId)
                    else:
                        isLowerBoundFound = True
                        if node.getAttribute(SbmlDef.VALUE) != "":
                            try:
                                self._tmpReaction.lb = \
                                    float(node.getAttribute(SbmlDef.VALUE))
                                isLowerBoundRead = True
                            except ValueError:
                                self._errorIllegalFloat(SbmlDef.P_LOWER_BOUND,
                                                       "reaction", reactionId)

                            if self._tmpReaction.lb <= -self.infLevel:
                                self._tmpReaction.lb = -inf
                        else:
                            self._errorMissingValue(SbmlDef.P_LOWER_BOUND,
                                                    "reaction", reactionId)

                elif nodeId == ubDefString or nodeName == ubDefString:
                    if isUpperBoundFound:
                        self._warnDuplicateNode(SbmlDef.P_UPPER_BOUND,
                                                "reaction", reactionId)
                    else:
                        isUpperBoundFound = True
                        if node.getAttribute(SbmlDef.VALUE) != "":
                            try:
                                self._tmpReaction.ub = \
                                    float(node.getAttribute(SbmlDef.VALUE))
                                isUpperBoundRead = True
                            except ValueError:
                                self._errorIllegalFloat(
                                                       SbmlDef.P_UPPER_BOUND,
                                                       "reaction", reactionId)

                            if self._tmpReaction.ub >= self.infLevel:
                                self._tmpReaction.ub = inf
                        else:
                            self._errorMissingValue(SbmlDef.P_UPPER_BOUND,
                                                    "reaction", reactionId)

                elif nodeId == fluxDefString or nodeName == fluxDefString:
                    if isFluxValFound:
                        self._warnDuplicateNode(SbmlDef.P_FLUX_VALUE,
                                                "reaction", reactionId)
                    else:
                        isFluxValFound = True
                        if node.getAttribute(SbmlDef.VALUE) != "":
                            try:
                                self._tmpReaction.fluxVal = \
                                    float(node.getAttribute(SbmlDef.VALUE))
                                isFluxValRead = True
                            except ValueError:
                                self._errorIllegalFloat(SbmlDef.P_FLUX_VALUE,
                                                       "reaction", reactionId)
                        else:
                            self._errorMissingValue(SbmlDef.P_FLUX_VALUE,
                                                    "reaction", reactionId)

                elif nodeId == objCoefDefString or nodeName == objCoefDefString:
                    if isObjCoefFound:
                        self._warnDuplicateNode(SbmlDef.P_OBJ_COEF,
                                                "reaction", reactionId)
                    else:
                        isObjCoefFound = True
                        if node.getAttribute(SbmlDef.VALUE) != "":
                            try:
                                self._tmpReaction.objCoef = \
                                    float(node.getAttribute(SbmlDef.VALUE))
                                isObjCoefRead = True
                            except ValueError:
                                self._errorIllegalFloat(SbmlDef.P_OBJ_COEF,
                                                       "reaction", reactionId)
                        else:
                            self._errorMissingValue(SbmlDef.P_OBJ_COEF,
                                                    "reaction", reactionId)

                elif nodeId == "":
                    msg = ("Warning: Found parameter without attribute \"%s\" "
                           "in %s of reaction '%s'." %
                           (SbmlDef.ID, SbmlDef.LISTOFPARAMETERS, reactionId))
                    self.messages.append((Verbosity.WARNING, msg))

                else:
                    msg = ("Debug: Ignoring parameter %s in %s of reaction '%s'"
                           "." % (nodeIdOrig, SbmlDef.LISTOFPARAMETERS,
                                  reactionId))
                    self.messages.append((Verbosity.DEBUG, msg))

            elif node.nodeType == dom.Node.ELEMENT_NODE and node.nodeName != "":
                # Anything that is not a parameter is ignored
                self._debugIgnoringNode(node.nodeName, "%s of reaction '%s'" %
                                        (SbmlDef.LISTOFPARAMETERS, reactionId))

        if not isLowerBoundFound:
            msg = ("Debug: No %s given for reaction '%s'. Set to default: "
                   "-inf." % (SbmlDef.P_LOWER_BOUND, reactionId))
            self.messages.append((Verbosity.DEBUG, msg))
        if not isUpperBoundFound:
            msg = ("Debug: No %s given for reaction '%s'. Set to default: "
                   "inf." % (SbmlDef.P_UPPER_BOUND, reactionId))
            self.messages.append((Verbosity.DEBUG, msg))

        success = ((not isLowerBoundFound or isLowerBoundRead) and
                   (not isUpperBoundFound or isUpperBoundRead) and
                   (not isFluxValFound or isFluxValRead) and
                   (not isObjCoefFound or isObjCoefRead))

        if (self._tmpReaction.lb > self._tmpReaction.ub):
            msg = ("Warning: %s (%r) > %s (%r) in reaction '%s'. If not fixed, "
                   "this may lead to undefined behavior in simulations." %
                   (SbmlDef.P_LOWER_BOUND, self._tmpReaction.lb,
                    SbmlDef.P_UPPER_BOUND, self._tmpReaction.ub, reactionId))
            self.messages.append((Verbosity.WARNING, msg))

        elif isFluxValRead:
            if (self._tmpReaction.fluxVal < self._tmpReaction.lb or
                self._tmpReaction.fluxVal > self._tmpReaction.ub):
                msg = ("Warning: %s of reaction '%s' (%r) is out of bounds "
                       "(%r, %r)." % (SbmlDef.P_FLUX_VALUE, reactionId,
                       self._tmpReaction.fluxVal, self._tmpReaction.lb,
                       self._tmpReaction.ub))
                self.messages.append((Verbosity.WARNING, msg))

        return success


class CacheReaction(object):
    """ Class for all collected information about a reaction
    """
    def __init__(self):
        self.clear()

    def addReactant(self, reactant):
        self.reactants.append(reactant)

    def addProduct(self, product):
        self.products.append(product)

    def clear(self):
        self.reactionName = ""
        self.reactionId = ""
        self.lb = -inf
        self.ub = inf
        self.reversible = 1
        self.reactants = []
        self.products = []
        self.notes = ""
        self.fluxVal = 0.
        self.objCoef = 0.


def main():
    # 1. Parse command line

    usage = "Usage: %prog [options]"
    version = "SBML parser\n" + COPYRIGHT_VERSION_STRING
    parser = OptionParser(usage=usage, version=version)

    parser.add_option("-f", "--file", dest="sbmlFile", help="SBML FILE to "
                      "be converted to metano format", metavar="FILE")
    parser.add_option("-o", "--output", dest="outputFile", help="write "
                      "extracted reactions to FILE", metavar="FILE")
    parser.add_option("-p", "--param", dest="outputParamFile", help="write "
                      "extracted parameters to scenario FILE", metavar="FILE")
    parser.add_option("-q", "--flux", dest="outputFluxFile", help="write fluxes"
                      " (if present) to FILE", metavar="FILE")
    parser.add_option("-i", "--inf-above", dest="infLevel", type="float",
                      help="set LEVEL above which a bound is considered "
                      "infinite (default 1000)", metavar="LEVEL")
    parser.add_option("-n", "--usenames", dest="usenames",
                      help="use metabolite and reaction names instead of IDs",
                      action="store_true")
    parser.add_option("-s", "--suppress-comments", dest="noComments",
                      help="do not write comments to reaction file",
                      action="store_true")
    parser.add_option("-x", "--solver", dest="solver", help="write SOLVER for "
                      "FBA to the scenario file", metavar="SOLVER")
    parser.add_option("-b", "--boundary-regex", dest="boundaryRegex",
                      help="consider all metabolites whose IDs match regular "
                      "expression PATTERN outside the system boundary (in "
                      "addition to metabolites with boundaryCondition=\"true\")"
                      " -- ^xxx matches at beginning, xxx$ at end",
                      metavar="PATTERN")
    parser.add_option("-d", "--debug-info", dest="debugInfo", help="print"
                      " debug information", action="store_true")

    parser.set_defaults(infLevel=1000., debugInfo=False, usenames=False,
                        noComments=False)

    options, _ = parser.parse_args()
    parser.check_required("-f")
    parser.check_required("-p")
    parser.check_required("-o")

    # 2. Set infinity level for bounds

    if options.infLevel <= 0.:
        print("Error: Infinity level must be positive")
        exit()
    sbmlP = SbmlParser(options.infLevel)

    # 3. Read the SBML file

    print("Start parsing..")
    print("SBML file             '%s'" % options.sbmlFile)
    print("Output reaction file: '%s'" % options.outputFile)
    print("Output scenario file: '%s'" % options.outputParamFile)
    if options.boundaryRegex:
        print("Boundary reg-ex:      '%s'" % options.boundaryRegex)
    print("Infinity starts at:   %r\n" % options.infLevel)

    try:
        success = sbmlP.readFile(options.sbmlFile, options.boundaryRegex)
    except ExpatError as strerror:
        print("Error in file %s:" % os.path.basename(options.sbmlFile))
        print(strerror)
        exit()
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.sbmlFile))
        print(strerror)
        exit()

    if options.debugInfo:
        msgs = sbmlP.getMessages(Verbosity.DEBUG)
    else:
        # Show only error, warning, and info messages
        msgs = sbmlP.getMessages()
    if msgs:
        print(('\n'.join(msgs)).encode('ascii', 'replace'))

    # Do not try to build the model if a critical error occurred
    if not success:
        exit()

    # 4. Replace illegal characters (whitespace, :, #) with underscores

    sbmlP.fixNames()

    # 5. Build the model

    model = sbmlP.buildModel(options.usenames)

    if options.usenames:
        # Get unique reaction labels
        reaLabels = dict(list(zip(model.getReactionNames(),
                             model._getTranslatedReactionNames())))
    else:
        reaLabels = None

    # 6. Construct objective function string

    objCoefficients = sbmlP.getObjCoefficients(reaLabels)
    objStr = ""

    if isZeroVector(list(objCoefficients.values())):
        # No coefficients found - look for 'biomass' reaction instead

        # Make a regex pattern for case-insensitive matching of 'biomass'
        pattern = "".join("[%c%c]" % (c, c.lower()) for c in
                          SbmlDef.BIOMASS.upper())
        biomassReactions = [x.name for x in model.getSubmodelByRegex(pattern)]
        nBiomassReactions = len(biomassReactions)
        if nBiomassReactions >= 1:
            objStr = biomassReactions[0]
            if nBiomassReactions > 1:
                print ("Warning: More than one biomass reaction: " +
                       ", ".join(biomassReactions))
        else:
            print ("Warning: No biomass reaction found! (OBJ line must be "
                   "added manually to scenario file.)")
    else:
        # Create objective function string from vector
        for rea in objCoefficients:
            objStr += makeCoefNameTerm(objStr, objCoefficients[rea], rea)

    # 7. Write reaction file

    try:
        model.writeToFile(options.outputFile, not options.noComments)
    except IOError as strerror:
        print ("Unable to write to file %s:" %
               os.path.basename(options.outputFile))
        print(strerror)
        exit()

    # 8. Write scenario file

    try:
        with open(options.outputParamFile,"w") as f:
            preamble = ""
            if objStr:
                preamble += "OBJ max %s\n" % objStr
            if options.solver:
                preamble += "SOLVER %s\n" % options.solver
            f.write(preamble + "\n")
            model.writeBoundsToParamFileHandle(f)
    except IOError as strerror:
        print ("Unable to write to file %s:" %
               os.path.basename(options.outputParamFile))
        print(strerror)
        exit()

    # 9. Write flux file

    if options.outputFluxFile:
        fluxDict = sbmlP.getFluxValues(reaLabels)
        rea_bounds = [(rea.name, (rea.lb, rea.ub)) for rea in model]
        if options.usenames:
            # Replace IDs with names in boundsDict
            boundsDict = {}
            for rea, bounds in rea_bounds:
                boundsDict[reaLabels[rea]] = bounds
        else:
            boundsDict = dict(rea_bounds)

        flux = MetabolicFlux(fluxDict, boundsDict)
        try:
            flux.writeToFile(options.outputFluxFile)
        except IOError as strerror:
            print ("Unable to write to file %s:" %
                   os.path.basename(options.outputFluxFile))
            print(strerror)
            exit()

    print("Finished successfully.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python

""" Script for determining the metabolites in active reactions

This script parses an FBA solution file and the underlying reaction file and
prints all metabolites present in reactions with a non-zero flux along with
the number of such reactions in which they appear. It also lists the active
reactions.


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function
from __future__ import absolute_import

from metano.fba import OptionParser
from metano.reactionparser import ReactionParser
from metano.metabolicmodel import MetabolicModel
from metano.metabolicflux import MetabolicFlux
from metano.defines import COPYRIGHT_VERSION_STRING
import os


def main():
    # 1. Parse command line

    usage = "Usage: %prog [options]"
    version = "%prog\n" + COPYRIGHT_VERSION_STRING
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-r", "--reactions", dest="reactionFile",
                      help="use the given reaction FILE", metavar="FILE")
    parser.add_option("-s", "--solution", dest="solutionFile",
                      help="parse the given FBA solution FILE", metavar="FILE")
    parser.add_option("-c", "--cutoff", dest="cutoff", type="float",
                      help="threshold below which a flux is to be considered "
                      "zero; default: 1E-10)", metavar="VALUE")
    parser.set_defaults(cutoff=1E-10)

    options, _ = parser.parse_args()
    parser.check_required("-r")
    parser.check_required("-s")

    # 2. Parse reaction file

    rparser = ReactionParser()
    model = MetabolicModel()
    try:
        model.addReactionsFromFile(options.reactionFile, rparser)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in reaction file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()

    # 3. Parse solution file

    solution = MetabolicFlux()
    try:
        solution.readFromFile(options.solutionFile)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.solutionFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in solution file %s:" %
               os.path.basename(options.solutionFile))
        print(strerror)
        exit()

    # 4. Count metabolites' participation in active reactions

    model_messages = []
    activeReactions, metabolitesSolution = \
        model.getActiveReactions(solution.getVecOrderedByModel(model),
                                 options.cutoff, model_messages)

    # Show warning and info messages (if any)
    msgs = rparser.getMessages() + [x[1] for x in model_messages]
    if msgs:
        print('\n'.join(msgs)+'\n')

    # 5. Output results

    print("Active metabolites:\n")
    for met in sorted(metabolitesSolution, key=metabolitesSolution.get,
                      reverse=True):
        print(met, metabolitesSolution[met])

    print("\nActive reactions:\n")
    for reactionName in activeReactions:
        print(reactionName)

    print ("\n%u metabolites in %u active reactions." %
           (len(metabolitesSolution), len(activeReactions)))


if __name__ == "__main__":
    main()



"""
This module defines an iterator that returns a given number of random vectors.


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""

from builtins import object
from numpy.random import uniform


class StartPointIterator(object):
    """ Iterator for generating random start points, i.e. vectors with dimension
        n and entries from a uniform distribution in [lb, ub)
    """
    def __init__(self, dimension, numIterations):
        self.dim = dimension
        self.numIter = numIterations
        self.lb = -1.
        self.ub = 1.

    def setDimension(self, dimension):
        self.dim = dimension

    def setRange(self, lb=None, ub=None):
        if lb != None:
            self.lb = lb
        if ub != None:
            self.ub = ub

    def reset(self, numIterations):
        self.numIter = numIterations

    def __iter__(self):
        return self

    def __next__(self):
        if self.numIter == 0:
            raise StopIteration
        self.numIter -= 1
        return uniform(self.lb, self.ub, self.dim)



#!/usr/bin/env python
""" Script for listing all metabolite names

This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.

"""
from __future__ import print_function
from __future__ import absolute_import

from metano.metabolicmodel import MetabolicModel
import sys
import os

try:
    filename = sys.argv[1]
except IndexError:
    print("Error: No filename given.")
    print("Usage is\n    " + os.path.basename(sys.argv[0]), "<reaction file>")
    exit()

model = MetabolicModel()
try:
    model.addReactionsFromFile(filename)
except IOError as strerror:
    print ("An error occurred while trying to read file %s:" %
           os.path.basename(filename))
    print(strerror)
    exit()
except SyntaxError as strerror:
    print ("Error in reaction file %s:" %
           os.path.basename(filename))
    print(strerror)
    exit()
print("\n".join(sorted(model.getMetaboliteNames())))



#!/usr/bin/env python

"""
Script for adding bounds imposed by the irreversibility of some reactions to
a metano scenario file

This script reads a scenario file and a reaction file and writes a scenario
file. While the two scenario files can be the same, it is advisable to keep
the original scenario file, which is otherwise overwritten.

The central function is performed by
MetabolicModel.writeBoundsToParamFileHandle().


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function
from __future__ import absolute_import

from metano.fba import OptionParser
from metano.reactionparser import ReactionParser
from metano.paramparser import ParamParser
from metano.metabolicmodel import MetabolicModel
from metano.defines import COPYRIGHT_VERSION_STRING
import os


def main():
    # 1. Parse command line

    usage = "Usage: %prog [options]"
    version = "%prog\n" + COPYRIGHT_VERSION_STRING
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-p", "--parameters", dest="inputFile",
                      help="use the given scenario FILE", metavar="FILE")
    parser.add_option("-r", "--reactions", dest="reactionFile", help="use "
                      "reversibilities from reaction FILE", metavar="FILE")
    parser.add_option("-o", "--output", dest="outputFile", help="write output "
                      "(scenario file) to FILE", metavar="FILE")

    options, _ = parser.parse_args()
    parser.check_required("-p")
    parser.check_required("-r")
    parser.check_required("-o")

    # 2. Parse reaction file

    rparser = ReactionParser()
    model = MetabolicModel()
    try:
        model.addReactionsFromFile(options.reactionFile, rparser)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in reaction file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()

    # 3. Parse scenario file

    model_messages = []
    pparser = ParamParser()
    try:
        # Parse file, get maxmin, name of objective function, and solver name
        maxmin, obj_name, solver, numIter, lb, ub = \
            pparser.parse(options.inputFile)
        # Set flux bounds in model (for parameter check)
        model.setFiniteBounds(lb, ub, True, model_messages)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.inputFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in scenario file %s:" %
               os.path.basename(options.inputFile))
        print(strerror)
        exit()
    except ValueError as strerror:
        print(strerror)
        exit()

    # Show warning and info messages of parsers
    msgs = (rparser.getMessages() + pparser.getMessages() +
            [x[1] for x in model_messages])
    if msgs:
        print('\n'.join(msgs))

    # 4. Write new scenario file

    try:
        with open(options.outputFile, 'w') as f:
            str_maxmin = {True: "max", False: "min"}[maxmin]
            f.write("OBJ %s %s\n" % (str_maxmin, obj_name))
            if solver != "":
                f.write("SOLVER %s\n" % solver)
            if numIter >= 0:
                f.write("NUM_ITER %u\n" % numIter)
            f.write("\n")
            model.writeBoundsToParamFileHandle(f)
            f.write("\n")
    except IOError as strerror:
        print ("An error occurred while trying to write file %s:" %
               os.path.basename(options.outputFile))
        print(strerror)
        exit()


if __name__ == "__main__":
    main()



from __future__ import print_function
from builtins import str
from builtins import map
from builtins import object
import sys
from itertools import chain
from pprint import pprint

import MySQLdb
from _mysql_exceptions import OperationalError
import networkx as nx

def flatten(l):
    return list(chain.from_iterable(l))

def listToString(l):
    return "', ".join("'" + str(i) for i in l) + "'"

class MetabolicMap(object):

    reactionShape = "ellipse"
    reactionColor = "#ffcd00"
    
    substanceShape = "roundrectangle"
    substanceColor = "#89a400"
    
    edgeWidth = "3"
    edgeColor =	"#000000"
    edgeArrow = "standard"
    
    fontSize = "14"

    # TODO faded colors    
#    reactionColor = "#ffcd0030"
#    substanceColor = "#89a40030"

    def __init__(self, server, user, password, database, verbosityLevel=0):
        
        self.server = server
        self.user = user
        self.__password = password
        self.database = database
        self.verbosityLevel = verbosityLevel
        self.__connection = None
        self.clearGraph()
    
    def __del__(self):
        self.closeConnection()
    
    def __getConnection(self):
        
        if self.__connection is None: # TODO check for timeout
            # open new db connection
            self.__connection = MySQLdb.connect(self.server, self.user,
                                                self.__password, self.database)
        return self.__connection

    def closeConnection(self):
        
        if self.__connection is not None:
            try:
                # close db connection
                self.__connection.commit()
                self.__connection.close()
                del self.__connection
                self.__connection = None
            except OperationalError:
                # connection is already closed
                del self.__connection
                self.__connection = None

    def clearGraph(self):
        self.graph=nx.DiGraph()

    def printStats(self):
            print("graph has",self.graph.number_of_nodes(),"nodes and", \
                  self.graph.number_of_edges(),"edges")    

    # TODO try other export formats
    def writeToGML(self, filename):
        if self.verbosityLevel > 0:
            self.printStats()
        nx.write_gml(self.graph, filename)
        
#    def writeToDot(self, filename):
#        if self.verbosityLevel > 0:
#            self.printStats()
#        nx.write_dot(self.graph, filename)

#    def writeToGraphML(self, filename):
#        # TODO fails if there are no edges
#        if self.verbosityLevel > 0:
#            self.printStats()
#        nx.write_graphml(self.graph, filename)

    def drawReactionNodes(self, reactionNodeList):

        connection = self.__getConnection()
        cursor = connection.cursor()

        reactionNodeString = listToString(reactionNodeList)
        
        # add all edges and nodes to the graph
        
        cursor.execute("""
        SELECT DISTINCT edge.source_node_id, edge.target_node_id
        FROM edge
        WHERE (edge.source_node_id IN (%s)
        OR edge.target_node_id IN (%s))
        """ % (reactionNodeString, reactionNodeString))
        
        result = cursor.fetchall()
        
        if len(result)<=0:
            print("No edges to draw.")
            return

        self.graph.add_edges_from(result)

        # assign edge attributes
        for start, end  in self.graph.edges_iter():

            # skip already existing edges
            if "graphics" in self.graph[start][end]:
                continue
            
            self.graph[start][end]["graphics"]= \
                { "width":float(self.edgeWidth),
                 "fill":self.edgeColor,
                 "targetArrow":self.edgeArrow
                 }

        # list of all node ids as string
        nodeString = listToString(flatten(result))
        
        # get position, width and height of each node
        cursor.execute("""
        SELECT node_id, x_pos, y_pos
        FROM node
        WHERE node_id IN (%s)
        """ % nodeString)

        result = cursor.fetchall()
        positionDict = dict( (nodeID, (x, y)) for nodeID, x, y in result)

        # get EC numbers as labels
        cursor.execute("""
        SELECT enzyme2node.node_id, enzyme.ec_number
        FROM enzyme2node INNER JOIN enzyme ON (enzyme.enzyme_id=enzyme2node.enzyme_id)
        WHERE enzyme2node.node_id IN (%s)
        """ % nodeString)
        result = cursor.fetchall()
        
        # assign reaction node attributes
        for nodeID, label in result:
            
            # skip already existing nodes
            if "graphics" in self.graph.node[nodeID]:
                continue
            
            self.graph.node[nodeID]["label"]=label
            self.graph.node[nodeID]["graphics"]= \
                {"x":positionDict[nodeID][0],
                 "y":positionDict[nodeID][1],
                 "w":max(30.0, len(label)*9.), # TODO magic numbers
                 "h":30.0,
                 "type":self.reactionShape,
                 "fill":self.reactionColor
                 }
            self.graph.node[nodeID]["LabelGraphics"] = \
                {"fontSize":int(self.fontSize)}

        # get substance names as labels
        cursor.execute("""
        SELECT substance2node.node_id, substance.name
        FROM substance2node INNER JOIN substance ON (substance.substance_id=substance2node.substance_id)
        WHERE substance2node.node_id IN (%s)
        """ % nodeString)
        result = cursor.fetchall()

        # assign substance nodes attributes        
        for nodeID, label in result:
            
            # skip already existing nodes
            if "graphics" in self.graph.node[nodeID]:
                continue
            
            self.graph.node[nodeID]["label"]=label
            self.graph.node[nodeID]["graphics"]= \
                {"x":positionDict[nodeID][0],
                 "y":positionDict[nodeID][1],
                 "w":max(30.0, len(label)*9.), # TODO magic numbers
                 "h":30.0,
                 "type":self.substanceShape,
                 "fill":self.substanceColor
                 }
            self.graph.node[nodeID]["LabelGraphics"] = \
                {"fontSize":int(self.fontSize)}
        

    # TODO MetaCyc only
    def drawReactions(self, reactionIDlist):
        
        connection = self.__getConnection()
        cursor = connection.cursor()
        
        # translate reaction IDs to enzyme nodes
        
        reactionString = listToString(reactionIDlist)
        
        cursor.execute("""
        SELECT r2n.node_id
        FROM reaction2kegg_metacyc_brenda AS r2db
        INNER JOIN reaction2enzymenode AS r2n
        ON (r2db.reaction_id=r2n.reaction_id)
        WHERE Reaction_ID_MetaCyc IN (%s)
        """ % reactionString)
        
        result = cursor.fetchall()

        self.drawReactionNodes(list(map(str,flatten(result))))
    
    def drawEnzymes(self, ecList):
        
        connection = self.__getConnection()
        cursor = connection.cursor()
        
        ecString = listToString(ecList)
        
        # translate EC numbers to enzyme nodes
        
        cursor.execute("""
        SELECT enzyme2node.node_id
        FROM enzyme INNER JOIN enzyme2node ON (enzyme.enzyme_id=enzyme2node.enzyme_id)
        WHERE enzyme.ec_number IN (%s)
        """ % ecString)
        
        result = cursor.fetchall()
        
        self.drawReactionNodes(list(map(str,flatten(result))))
        
        # TODO return list of drawn enzymes
    
    def drawReactionEnzyme(self, reactionEClist):

        connection = self.__getConnection()
        cursor = connection.cursor()
            
        reactionString = listToString(rid for rid, _ in reactionEClist)
        ecString = listToString(ec for _, ec in reactionEClist if ec is not None)
        
        cursor.execute("""
        -- translate reaction id to node id
        SELECT r2n.node_id
        FROM reaction2kegg_metacyc_brenda AS r2db
        INNER JOIN reaction2enzymenode AS r2n
        ON (r2db.reaction_id=r2n.reaction_id)

        -- translate node id to ec numbers
        INNER JOIN enzyme2node AS e2n
        ON (r2n.node_id=e2n.node_id)
        INNER JOIN enzyme AS e
        ON (e2n.enzyme_id=e.enzyme_id)
        
        -- reactions from model, allow only ECs from model
        WHERE Reaction_ID_MetaCyc IN (%s)
        AND e.ec_number IN (%s)
        """ % (reactionString, ecString))
        
        result = cursor.fetchall()
        
        if len(result)<=0:
            print("No reaction could be mapped.")
            return
        
        self.drawReactionNodes(list(map(str,flatten(result))))
        
        # find untranslated reaction ids
        cursor.execute("""
        SELECT r2db.Reaction_ID_MetaCyc
        FROM reaction2kegg_metacyc_brenda AS r2db
        """)
        
        result = cursor.fetchall()
        allDBids = set(flatten(result))
        missingIDs = set(rid for rid, _ in reactionEClist).difference(allDBids)
#        pprint(missingIDs)
        
        reactionIDtoECdict = dict((r,e) for r,e in reactionEClist if e is not None and r in missingIDs)
        
#        pprint(reactionIDtoECdict)
        self.drawEnzymes(list(reactionIDtoECdict.values()))
        
        # TODO return list of drawn reactions

    def drawRemainingReactions(self, alpha="30"): # alpha unused
        
        connection = self.__getConnection()
        cursor = connection.cursor()

        if len(self.graph)<=0:
            print("The graph is empty.")
            return

        nodeString = listToString(self.graph.nodes())
        
        # add all edges and nodes to the graph
        
        cursor.execute("""
        SELECT DISTINCT node_id
        FROM enzyme2node
        WHERE node_id NOT IN (%s)
        """ % nodeString)
        
        result = cursor.fetchall()
        
        # TODO set as parameter(object?)
        self.reactionColor = "#ffcd0030"
        self.substanceColor = "#89a40030"
        self.edgeColor = "#00000030"
        
        self.drawReactionNodes(flatten(result))

    def drawPathwayLabels(self):

        connection = self.__getConnection()
        cursor = connection.cursor()

        # draw pathway frame nodes

#        cursor.execute("""
#            SELECT MIN(node_id) AS node_id, AVG(x_pos) AS x, AVG(y_pos) AS y,
#            (MAX(x_pos)-MIN(x_pos)) AS width, (MAX(y_pos)-MIN(y_pos)) AS height
#            FROM `node`
#            WHERE `node_type` LIKE 'pathwayFrame'
#            GROUP BY `dominant_pathway_id`
#            """)
#        result = cursor.fetchall()
#        
#        for nodeID, x, y, width, height in result:
#            self.graph.add_node(nodeID)
#            self.graph.node[nodeID]["label"]=""
#            self.graph.node[nodeID]["graphics"]= \
#                {"x":x,
#                 "y":y,
#                 "w":width,
#                 "h":height,
#                 "hasFill":"0",
#                 "outlineStyle": "\"dashed\"",
#                 "type":"\"rectangle\"",
#                 }

        # draw pathway label nodes
        # TODO increase font size
        # TODO color

        cursor.execute("""
            SELECT node_id, node_label, x_pos, y_pos, width, height
            FROM node
            WHERE `node_type` LIKE 'Pathway_title'
            """)
            
        result = cursor.fetchall()
        
        for nodeID, label, x, y, width, height in result:
            self.graph.add_node(nodeID)
            self.graph.node[nodeID]["label"]=label
            self.graph.node[nodeID]["graphics"]= \
                {"x":x,
                 "y":float(y)-0.5*30., # move labels
                 "w":max(30.0, len(label)*9.), # TODO magic numbers
                 "h":30.0,
                 "type":self.substanceShape,
#                 "fill":"\""+self.reactionColor+"\""
                 }

    def setNodeGraphicsAttribute(self, attribute, valueList, attributeDict):
        """
        Set all graphics attributes in attributeDict of all nodes in the graph
        whose value of the given attribute is in the valueList
        
        @param attribute: The nodes will be filtered by this attribute
        @type attribute: string
        @param valueList: List of values to match with the attribute
        @type valueList: list
        @param attributeDict: Dict of GML graphics attributes
        @type attributeDict: dict
        
        returns a subset of valueList with attributes which could _not_ be
        found in any node of the graph.
        """
        # TODO example

        foundAttributes = set()
        
        for node in self.graph.nodes_iter():

            if self.graph.node[node][attribute] in valueList:

                if self.verbosityLevel>1:
                    print(node, "with attribute", \
                          self.graph.node[node][attribute], "found")
                
                foundAttributes.add(self.graph.node[node][attribute])              
            
                for key, value in attributeDict.items():
                    self.graph.node[node]["graphics"][key]=value

        if self.verbosityLevel>0:
            print(len(foundAttributes), "of", len(set(valueList)), "distinct", \
                  "attributes found")

        return set(valueList).difference(foundAttributes)


def main(argv=None):

    if argv is None:
        argv = sys.argv

    server = "fileserver.bioinfo.nat.tu-bs.de"
    user = "rre"
    password = open("pw.txt").readline().strip()
    database = "metabolic_pathways"
    verbosityLevel = 1

#    mymap = MetabolicMap(server, user, password, database, verbosityLevel)
#    mymap.drawEnzymes(["6.4.1.1", "1.1.1.35"])
#    mymap.writeToGML("/home/rre/Desktop/test.gml")
#    mymap.writeToDot("/home/rre/Desktop/test.dot")
#    mymap.writeToGraphML("/home/rre/Desktop/test.graphml")
#    mymap.closeConnection()
    
    connection = MySQLdb.connect(server, user,password,"rre_roseobacter")
    cursor = connection.cursor()
    
    megateriumMap = MetabolicMap(server, user, password, database, verbosityLevel)
    
#    cursor.execute("""
#    SELECT DISTINCT Bmegaterium__model.ec
#    FROM Bmegaterium__model
#    INNER JOIN Bmegaterium__scenario
#    ON (Bmegaterium__model.reaction_id=Bmegaterium__scenario.reaction_id)""")
#    result = cursor.fetchall()
#    ecList = flatten(result)
#    megateriumMap.drawEnzymes(ecList)

#    cursor.execute("""
#    SELECT DISTINCT reaction_id
#    FROM Bmegaterium__scenario
#    """)
#    result = cursor.fetchall()
#    reactionIDlist = flatten(result)
#    megateriumMap.drawReactions(reactionIDlist)

    cursor.execute("""
    SELECT reaction_id, ec
    FROM Bmegaterium__table_BiomassGFP
    """)
    reactionEClist = cursor.fetchall()
    megateriumMap.drawReactionEnzyme(reactionEClist)
    megateriumMap.drawRemainingReactions()
    megateriumMap.drawPathwayLabels()

    # TODO metabolite liste einlesen
#    metabList=list(open("/home/rre/Desktop/Megaterium/BmegateriumT2.csv"))
#    metabList=[m.strip() for m in metabList]
#    notFound = megateriumMap.setNodeGraphicsAttribute("label",metabList,{"fill":"\"#CC0000\""})
#    print notFound
    

    megateriumMap.writeToGML("/home/rre/Desktop/Bmega.gml")
    
    megateriumMap.closeConnection()
    connection.close()
    
#    dinoMap = MetabolicMap(server, user, password, database, verbosityLevel)    
#    cursor.execute("""
#    SELECT reaction_id, ec
#    FROM `Dshibae__table_beta-D-glucose`
#    """)
#    reactionEClist = cursor.fetchall()
#    dinoMap.drawReactionEnzyme(reactionEClist)
#    dinoMap.drawPathwayLabels()
#    dinoMap.drawRemainingReactions()
#    dinoMap.writeToGML("/home/rre/Desktop/dino-glucose.gml")
#    dinoMap.closeConnection()
#    
#    dinoMap = MetabolicMap(server, user, password, database, verbosityLevel)
#    cursor.execute("""
#    SELECT reaction_id, ec
#    FROM Dshibae__model
#    """)
#    reactionEClist = cursor.fetchall()
#    dinoMap.drawReactionEnzyme(reactionEClist)
#    dinoMap.drawPathwayLabels()
#    dinoMap.drawRemainingReactions()
#    dinoMap.writeToGML("/home/rre/Desktop/dino-all.gml")
#    dinoMap.closeConnection()
    
#    phaeoMap = MetabolicMap(server, user, password, database, verbosityLevel)
#    cursor.execute("""
#    SELECT reaction_id, ec
#--    FROM `Pgallaeciensis__succinate_L-valine`
#    FROM `Pgallaeciensis__succinate_beta-D-glucose`
#    """)
#    reactionEClist = cursor.fetchall()
#    phaeoMap.drawReactionEnzyme(reactionEClist)
#    phaeoMap.drawPathwayLabels()
#    phaeoMap.drawRemainingReactions()
#    phaeoMap.writeToGML("/home/rre/Desktop/phaeo-glucose.gml")
#    phaeoMap.closeConnection()
    
    return 0

if __name__ == "__main__":

    sys.exit(main())



#!/usr/bin/env python
"""

This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function
from __future__ import absolute_import

from metano.fba import OptionParser
from metano.metabolicmodel import MetabolicModel
from metano.defines import COPYRIGHT_VERSION_STRING
import os
import sys


def main():

    # 1. Parse command line

    usage = "Usage: %prog <filename> [options]"
    version = "%prog\n" + COPYRIGHT_VERSION_STRING
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-c", "--compartment-suffixes",
                      dest="compartmentSuffixes", help="STRING of compartment "
                      "suffixes (separated by whitespace)", metavar="STRING")

    options, args = parser.parse_args()

    if len(args) < 1:
        print("Error: No filename given.")
        print ("Usage is\n    " + os.path.basename(sys.argv[0]) + " <filename> "
               "[options]")
        exit()

    filename = args[0]

    # 2. Parse reaction file

    model = MetabolicModel()
    try:
        model.addReactionsFromFile(filename)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(filename))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in reaction file %s:" %
               os.path.basename(filename))
        print(strerror)
        exit()

    # 3. Identify exchange reactions

    exchangeReactions = [reaction for reaction in model if len(reaction) == 1]
    nExchange = len(exchangeReactions)
    print("Exchange reactions:", nExchange)
    print("\n".join([r.name for r in exchangeReactions
                     if "exchange" not in r.name]))

    # 4. Identify transport reactions

    if not options.compartmentSuffixes:
        print ("Warning: No compartment suffixes given. Unable to identify "
               "transport reactions.")


if __name__ == "__main__":
    main()



"""
This module defines a number of objective functions for nonlinear optimization.


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import division

from builtins import range
from past.utils import old_div
from numpy import dot

CUTOFF = 1E-20
LAX_CUTOFF = 1E-8


def biomass_per_flux(flux_vec, biomass_index):
    """ nonlinear objective function biomass per flux unit
    """
    denominator = (flux_vec**2).sum()
    if denominator < CUTOFF:
        return 0.   # Punish all-zero flux (in this case, the biomass flux is
                    # very low as well)
    return old_div(flux_vec[biomass_index],denominator)


def neg_grad_biomass_per_flux(flux_vec, biomass_index):
    """ compute gradient of biomass_per_flux function
    """
    denominator = (flux_vec**2).sum()
    sqdenom = denominator*denominator
    if sqdenom < CUTOFF:
        return 0.   # Punish all-zero flux (in this case, the numerator is very
                    # low as well)
    factor = 2.*flux_vec[biomass_index]/sqdenom
    solution = flux_vec*factor
    solution[biomass_index] -= 1./denominator
    return solution


def linear_to_lambda_func(vec):
    """ transform the linear function given by the coefficient vector into a
        function object
    """
    nonzero_indices = []
    for i in range(len(vec)):
        if abs(vec[i]) >= LAX_CUTOFF:
            nonzero_indices.append(i)

    # Special case: only one non-zero index
    if len(nonzero_indices) == 1:
        i = nonzero_indices[0]
        return lambda x : x[i]*vec[i]

    return lambda x : dot(x, vec)



#!/usr/bin/env python

""" This script computes the split ratios for the fluxes entering and leaving
    each metabolite (as percentages).

Usage:

    splitratios.py <solution-file> -r reaction-file [-o output-file]
                   [-m metabolite] [-t threshold] [-u] [-c cutoff] [-f] [-a]


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function
from __future__ import absolute_import

from builtins import range
import os
import sys
import csv
from metano.fba import OptionParser
from metano.reactionparser import ReactionParser
from metano.metabolicmodel import MetabolicModel
from metano.metabolicflux import MetabolicFlux
from metano.defines import COPYRIGHT_VERSION_STRING


def main():
    # 1. Parse command line

    usage = "Usage: %prog <solution-file> [options]"
    version = "Split-ratio analysis\n" + COPYRIGHT_VERSION_STRING
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-r", "--reactions", dest="reactionFile", help="use "
                      "metabolic model given by reaction FILE", metavar="FILE")
    parser.add_option("-m", "--metabolite", dest="metabolite", help="NAME of "
                      "the single metabolite to be analyzed (optional)",
                      metavar="NAME")
    parser.add_option("-o", "--output", dest="outputFile", help="perform "
                      "analysis for all metabolites and write output to FILE",
                      metavar="FILE")
    parser.add_option("-t", "--omit-fluxes-below", dest="threshold",
                      type="float", help="omit metabolites in output whose flux"
                      " is below the given THRESHOLD (only with -o)",
                      metavar="THRESHOLD")
    parser.add_option("-u", "--omit-unbranched", dest="omitUnbranched",
                      action="store_true", help="if set, metabolites with "
                      "unbranched flux are omitted (only with -o)")
    parser.add_option("-c", "--cutoff", dest="cutoff", type="float", help="use "
                      "the given cutoff VALUE (default 0)", metavar="VALUE")
    parser.add_option("-f", "--cutoff-is-absolute", dest="cutoffIsAbsolute",
                      action="store_true", help="if set, cutoff is in absolute "
                      "flux units rather than a value between 0 and 1")
    parser.add_option("-a", "--list-all", dest="listAll", action="store_true",
                      help="if set, cutoff is ignored, and even zero fluxes are"
                      " shown (per definition as outgoing)")
    parser.set_defaults(threshold=-1., omitUnBranched=False, cutoff=0.,
                        cutoffIsAbsolute=False, listAll=False)

    options, args = parser.parse_args()
    parser.check_required('-r')

    if not options.metabolite and not options.outputFile:
        print("Neither metabolite nor output file given. Nothing to do.")
        exit()

    # 2. Read solution file

    solution = MetabolicFlux()
    try:
        solution.readFromFile(args[0])
    except IndexError:
        print ("Error: No solution file given.\nUsage is\n    " +
               os.path.basename(sys.argv[0]) + " <solution-file> [options]")
        exit()
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(args[0]))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("An error occurred parsing file %s:" %
               os.path.basename(args[0]))
        print(strerror)
        exit()

    # 3. Parse reaction file

    model = MetabolicModel()
    rparser = ReactionParser()
    try:
        model.addReactionsFromFile(options.reactionFile, rparser)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in reaction file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()

    # 4. Compute split ratios

    if options.outputFile:
        # Compute split ratios for all metabolites
        splitRatios = solution.computeAllSplitRatios(model, options.cutoff,
                                                     options.cutoffIsAbsolute, options.listAll)
        try:
            writeSplitRatiosToFile(options.outputFile, splitRatios,
                                   options.threshold, options.omitUnbranched)
        except IOError as strerror:
            print ("Unable to write to file %s:" %
                   os.path.basename(options.outputFile))
            print(strerror)

        # Print statistics
        nMetabolites = len(splitRatios)
        print ("The model contains %u reactions and %u metabolites." %
               (len(model), nMetabolites))
        EPSILON = 1e-6
        nZero = nOne = nBelowEps = nDiff = 0
        maxDiff = 0.
        maxdMet = ""

        for met in splitRatios:
            outRatios, inRatios = splitRatios[met]
            outSum = sum(outRatios[rea][1] for rea in outRatios)
            diffInOut = (abs(outSum - sum(inRatios[rea][1]
                                          for rea in inRatios)))
            if diffInOut > EPSILON:
                nDiff += 1
                if diffInOut > maxDiff:
                    maxDiff = diffInOut
                    maxdMet = met
            if len(outRatios) == 0 or outSum == 0.:
                nZero += 1
            else:
                if len(outRatios) == 1 and len(inRatios) == 1:
                    nOne += 1
                elif options.listAll:
                    # Actually count all outgoing and incoming fluxes > 0
                    nOut = sum(
                        1 for rea in outRatios if outRatios[rea][1] > 0.)
                    nIn = sum(1 for rea in inRatios if inRatios[rea][1] > 0.)
                    if nOut == nIn == 1:
                        nOne += 1
                if sum(outRatios[rea][1] for rea in outRatios) < EPSILON:
                    nBelowEps += 1

        print ("%u metabolites have a diff. betw. incoming and outgoing "
               "fluxes above %g.\nThe highest diff. betw. incoming and "
               "outgoing metabolite flux is %g (%s)." % (nDiff, EPSILON,
                                                         maxDiff, maxdMet))
        if options.cutoff == 0.:
            print ("%u metabolites (%.2g%%) have flux zero." %
                   (nZero, (float(nZero)/float(nMetabolites))*100.))
        else:
            if options.cutoffIsAbsolute:
                sCutoff = "%g" % options.cutoff
            else:
                sCutoff = "%g%%" % (options.cutoff * 100.)
            print ("%u metabolites (%.2g%%) have no single outgoing flux above "
                   "%s." % (nZero, (float(nZero)/float(nMetabolites))*100.,
                            sCutoff))
        if nBelowEps:
            print ("Another %u metabolites (%.2g%%) have a flux below %g." %
                   (nBelowEps, ((float(nBelowEps)/float(nMetabolites))*100.),
                    EPSILON))
        print ("%u metabolites (%.2g%%) lie on an unbranched path." %
               (nOne, ((float(nOne)/float(nMetabolites))*100.)))
        print()

        # Extract data for queried metabolite
        if options.metabolite:
            try:
                outRatios, inRatios = splitRatios[options.metabolite]
            except KeyError:
                print ("Error: Metabolite %r does not exist." %
                       options.metabolite)
                exit()

    elif options.metabolite:
        # Only compute split ratios for the given metabolite
        outRatios, inRatios = solution.computeSplitRatios(options.metabolite,
                                                          model, options.cutoff, options.cutoffIsAbsolute, options.listAll)

    if options.metabolite:
        print("Split ratios for %r" % options.metabolite)
        maxlen = max(len(max(outRatios, key=len)), len(max(inRatios, key=len)))
        fluxSum = sum(outRatios[rea][1] for rea in outRatios)
        print("- outgoing (sum %g) -" % fluxSum)
        for rea in sorted(outRatios, key=lambda x: outRatios[x][0],
                          reverse=True):
            ratio, flux = outRatios[rea]
            print (rea.ljust(maxlen) + " " + ("%.4g%%" % (ratio*100.)).rjust(10)
                   + " " + ("%.6g" % flux).rjust(12))
        fluxSum = sum(inRatios[rea][1] for rea in inRatios)
        print("- incoming (sum %g) -" % fluxSum)
        for rea in sorted(inRatios, key=lambda x: inRatios[x][0],
                          reverse=True):
            ratio, flux = inRatios[rea]
            print (rea.ljust(maxlen) + " " + ("%.4g%%" % (ratio*100.)).rjust(10)
                   + " " + ("%.6g" % flux).rjust(12))


def writeSplitRatiosToFile(filename, splitRatios, threshold, omitUnbranched):
    with open(filename, 'w') as f:
        if filename.split(".")[-1] == "csv":
            writeSplitRatiosToCSVHandle(f, splitRatios, threshold,
                                        omitUnbranched)
        else:
            writeSplitRatiosToFileHandle(f, splitRatios, threshold,
                                         omitUnbranched)


def writeSplitRatiosToCSVHandle(f, splitRatios, threshold, omitUnbranched):
    # sorted[Out/In]Keys stores the sorted keys for the outRatios or inRatios
    #  dict of each metabolite (in descending order of flux)
    sortedOutKeys, sortedInKeys = {}, {}
    # maxwidth stores maximum width of each column
    metaboliteList = sorted(splitRatios)
    csvwriter = csv.writer(f, delimiter=";", quotechar='"',
                           quoting=csv.QUOTE_MINIMAL)
    # Write Header to CSV file
    csvwriter.writerow(["NAME", "DIRECTION", "REACTION", "PERCENT", "FLUX"])

    for met in sorted(metaboliteList):
        outRatios, inRatios = splitRatios[met]

        if omitUnbranched and len(outRatios) <= 1 and len(inRatios) <= 1:
            # Omit metabolite in output if it has an unbranched flux
            continue

        if not outRatios:
            csvwriter.writerow([met, "outgoing", "", 0, 0])
        if not inRatios:
            csvwriter.writerow([met, "incoming", "", 0, 0])
        for o in sorted(outRatios, key=outRatios.get, reverse=True):
            csvwriter.writerow(
                [met, "outgoing", o, outRatios[o][0], outRatios[o][1]])
        for i in sorted(inRatios, key=inRatios.get, reverse=True):
            csvwriter.writerow(
                [met, "incoming", i, inRatios[i][0], inRatios[i][1]])


def writeSplitRatiosToFileHandle(f, splitRatios, threshold, omitUnbranched):
    # sorted[Out/In]Keys stores the sorted keys for the outRatios or inRatios
    #  dict of each metabolite (in descending order of flux)
    sortedOutKeys, sortedInKeys = {}, {}
    # maxwidth stores maximum width of each column
    maxwidth = {}
    metaboliteList = sorted(splitRatios)

    for met in metaboliteList:
        outRatios, inRatios = splitRatios[met]
        sortedOutKeys[met] = sorted(outRatios, key=lambda x: outRatios[x][0],
                                    reverse=True)
        sortedInKeys[met] = sorted(inRatios, key=lambda x: inRatios[x][0],
                                   reverse=True)
        for i in range(len(sortedOutKeys[met])):
            try:
                maxwidth[i] = max(maxwidth[i], len(sortedOutKeys[met][i]))
            except KeyError:
                maxwidth[i] = max(len(sortedOutKeys[met][i]), 12)
        for i in range(len(sortedInKeys[met])):
            try:
                maxwidth[i] = max(maxwidth[i], len(sortedInKeys[met][i]))
            except KeyError:
                maxwidth[i] = max(len(sortedInKeys[met][i]), 12)

    for met in metaboliteList:
        outRatios, inRatios = splitRatios[met]

        if omitUnbranched and len(outRatios) <= 1 and len(inRatios) <= 1:
            # Omit metabolite in output if it has an unbranched flux
            continue

        # 1st line: <metabolite name>
        # 2nd line: -outgoing <flux>
        # 3rd line: <rea1> <rea2> ... <reaN>
        # 4th line:  <%1>   <%2>  ...  <%N>
        # 5th line:  <v1>   <v2>  ...  <vN>
        # 6th line: -outgoing <flux>
        # 7th line: <rea1> <rea2> ... <reaN>
        # 8th line:  <%1>   <%2>  ...  <%N>
        # 9th line:  <v1>   <v2>  ...  <vN>
        lines = ["", "", ""]
        fluxSum = 0.
        for i in range(len(sortedOutKeys[met])):
            rea = sortedOutKeys[met][i]
            ratio, flux = outRatios[rea]
            fluxSum += flux
            lines[0] += " " + rea.rjust(maxwidth[i])
            lines[1] += " " + ("%.4g%%" % (ratio*100.)).rjust(maxwidth[i])
            lines[2] += " " + ("%.6g" % flux).rjust(maxwidth[i])

        if fluxSum < threshold:
            # Omit metabolite in output if its total flux is below threshold
            continue

        s = "%s\n-outgoing %g\n" % (met, fluxSum)
        for line in lines:
            if line:
                s += line + '\n'

        lines = ["", "", ""]
        fluxSum = 0.
        for i in range(len(sortedInKeys[met])):
            rea = sortedInKeys[met][i]
            ratio, flux = inRatios[rea]
            fluxSum += flux
            lines[0] += " " + rea.rjust(maxwidth[i])
            lines[1] += " " + ("%.4g%%" % (ratio*100.)).rjust(maxwidth[i])
            lines[2] += " " + ("%.6g" % flux).rjust(maxwidth[i])

        s += "-incoming %g\n" % fluxSum
        for line in lines:
            if line:
                s += line + '\n'
        f.write(s)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

""" This script computes the absolute difference between two flux distributions
    given as FBA solution files.

Usage:

    to_diff_solutions.py <file1> <file2> [-o output-file] [-c cutoff]

The output is written to the console if no output file is given.
Output format is as follows (without header):
<NAME>    <DIFFERENCE>

Reactions are sorted in order of decreasing flux difference. Reactions in which
the fluxes differ by less than cutoff are omitted.


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from builtins import map
from builtins import range
from past.utils import old_div
from metano.fba import OptionParser
from metano.metabolicflux import MetabolicFlux
from metano.defines import COPYRIGHT_VERSION_STRING
import os
import sys


def main():
    # 1. Parse command line

    usage = "Usage: %prog <file1> <file2> [options]"
    version = "%prog\n" + COPYRIGHT_VERSION_STRING
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-o", "--output", dest="outputFile", help="output FILE "
                      "(output is written to console if this is missing)",
                      metavar="FILE")
    parser.add_option("-c", "--cutoff", dest="cutoff", help="Cutoff below which"
                      " difference is considered zero; default is 1E-10")
    parser.set_defaults(cutoff="1E-10")

    options, args = parser.parse_args()

    if len(args) < 2:
        print("Error: Need two solution files.")
        print ("Usage is\n    " + os.path.basename(sys.argv[0]) + " <file1> "
               "<file2> [options]")
        exit()

    try:
        cutoff = float(options.cutoff)
    except ValueError:
        print("Error: Invalid floating point value for cutoff.")
        exit()
    if cutoff < 0.:
        print ("Warning: Cutoff is less than zero. Setting cutoff to zero, "
               "i.e. no cutoff.")
        cutoff = 0.

    basename = list(map(os.path.basename, args))
    # Use full path for identifying files if basenames are identical
    if basename[0] == basename[1]:
        basename = args

    # 2. Parse solution files and compute absolute differences

    solution = MetabolicFlux(), MetabolicFlux()
    for i in range(2):
        # Get pair of MetabolicFlux objects
        try:
            solution[i].readFromFile(args[i])
        except IOError as strerror:
            print ("An error occurred while trying to read file %s:" %
                   os.path.basename(basename[i]))
            print(strerror)
            exit()
        except SyntaxError as strerror:
            print("An error occurred parsing file %s:" % basename[i])
            print(strerror)
            exit()

    diffSolution = solution[0].absDiff(solution[1])

    # 3. Generate output

    nReactions = len(diffSolution)
    if nReactions == 0:
        exit()  # Nothing to do
    if len(diffSolution) == len(solution[0]) == len(solution[1]):
        print("Both solutions have the same %u reactions." % nReactions)
    else:
        n = list(map(len, solution))
        print ("The two solutions have %u reactions in common.\n(%s: %u, "
               "%s: %u)" % (n[0]+n[1]-nReactions, basename[0], n[0],
                            basename[1], n[1]))

    sumDiff = sum(diffSolution.fluxDict.values())
    sumSqDiff = sum(x*x for x in list(diffSolution.fluxDict.values()))
    print("Summed absolute difference: %s (mean: %s)" % (sumDiff,
                                                         old_div(sumDiff, nReactions)))
    print("Summed squared difference: %s (mean: %s)" % (sumSqDiff,
                                                        old_div(sumSqDiff, nReactions)))

    try:
        if options.outputFile is None:
            print()
            written = write_output(sys.stdout, diffSolution, solution, basename,
                                   cutoff, False)
        else:
            with open(options.outputFile, 'w') as f:
                written = write_output(f, diffSolution, solution, basename,
                                       cutoff, True)
    except IOError as strerror:
        print("An error occurred while trying to write output:")
        print(strerror)
        exit()

    if not written:
        print ("All differences are below the cutoff (%g). Nothing written." %
               cutoff)


def write_output(file_obj, diffSolution, solution, filenames, cutoff, showOrig):
    """ Write output to the given file object (expected to be open for writing)

    This function writes a list of "<reaction> <difference>" pairs to the given
    file object, sorted in order of decreasing difference. Differences below
    cutoff are not reported. If cutoff is zero, all differences are reported.
    If showOrig is True, the original flux values are listed in addition to the
    absolute differences.

    Keyword arguments:
    file_obj     -- the file object (open for writing; may also be sys.stdout)
    diffSolution -- MetabolicFlux of absolute flux differences
    solution     -- pair (or list) of solutions (flux distributions),
                    as MetabolicFlux
    filenames    -- pair (or list) of solution filenames, in original order
    cutoff       -- cutoff below which differences are not reported
    showOrig     -- if True, also show original values, else only the difference

    Returns:
    True if any difference is above cutoff, False else
    """

    # Get length of longest reaction name and length of filenames for formatting
    maxlen = len(max(list(diffSolution.fluxDict.keys()), key=len))
    len0 = max(len(filenames[0]), 12)
    len1 = max(len(filenames[1]), 12)

    first = True
    # Sort dictionary by value in decreasing order
    for rea in sorted(diffSolution.fluxDict, key=diffSolution.fluxDict.get,
                      reverse=True):
        value = diffSolution[rea]
        if value < cutoff:
            break

        if first:
            s = "REACTION".ljust(maxlen)+"DIFFERENCE".rjust(12)
            if showOrig:
                s += ('  '+filenames[0].rjust(len0)+'  ' +
                      filenames[1].rjust(len1))
            file_obj.write(s+'\n')
            first = False

        s = rea.ljust(maxlen) + ("%.6g" % value).rjust(12)
        in2 = rea in solution[1]
        if rea in solution[0]:
            if showOrig:
                s += '  '+("% .6g" % solution[0][rea]).rjust(len0)
            notPresent = -1
        else:
            notPresent = 0
            if showOrig and in2:
                s += ' '*(len0+2)
        if in2:
            if showOrig:
                s += '  '+("% .6g" % solution[1][rea]).rjust(len1)
        else:
            notPresent = 1

        if notPresent >= 0:
            s += " # not present in " + filenames[notPresent]
        file_obj.write(s+'\n')

    return not first


if __name__ == "__main__":
    main()




''' Performs batch computation of flux balance analysis

This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
'''
from __future__ import print_function
from builtins import zip
from builtins import str
from builtins import range
from builtins import object
import optparse
import os
from multiprocessing import Pool

from metano.metabolicmodel import MetabolicModel
from metano.reactionparser import ReactionParser
from metano.paramparser import ParamParser
from metano.fba import FbAnalyzer
from metano.defines import FbaParam, COPYRIGHT_VERSION_STRING


def unwrap_self_run(arg, **kwarg):
    """ snipped for running multiprocessing within a class
    from Thomas Rueckstiess:
    http://www.rueckstiess.net/research/snippets/show/ca1d7d90
    """
    return MultipleFBA.perform_fba(*arg, **kwarg)


class OptionParser(optparse.OptionParser):
    """ Extension of OptionParser with function for checking required arguments
    """

    def check_required(self, opt):
        """ check required arguments """
        if type(opt) == list:
            for i in opt:
                option = self.get_option(i)
                if getattr(self.values, option.dest) is not None:
                    print(i, getattr(self.values, option.dest))
                    return
            self.error(
                "you must provide at least on of these options: %s" % ", ".join(opt))

        option = self.get_option(opt)

        # Assumes the option's 'default' is set to None!
        if getattr(self.values, option.dest) is None:
            self.error("%s option not supplied" % option)


class MultipleFBA(object):
    """ Class for batch computation of FBA """

    def __init__(self, model, scenario, constraints, fba_file_prefix,
                 reactions_of_interest, verb=False):
        """ init """
        self.model = model
        self.verb = verb
        self.insce = []
        self.rates = {}
        self.run_count = 0
        self.fba_file_prefix = fba_file_prefix
        self.result = []
        self.roi = reactions_of_interest

        if constraints[0].split("\t")[0] != "NAME":
            print ("Error in file %s:"
                   "First line must begin with NAME flag!"
                   "See Metano documentation for further information")
            exit()
        self.names = [i.strip() for i in constraints[0].split("\t")[1:]]
        duplicates = set([x for x in self.names if self.names.count(x) > 1])
        if duplicates:
            print("Error: Found duplicate scenario name(s) in constraint matrix:")
            print("---", ", ".join(duplicates), "---")
            print("Scenario names MUST be unique!")
            exit()
        del constraints[0]

        for line in constraints:
            line = [i.strip() for i in line.split("\t")]
            if line[0] not in ("\t", "\n", "", " "):
                self.rates[line[0]] = [i if i else "0.0" for i in line[1:]]

        try:
            self.run_count = len(line)-1
        except UnboundLocalError:
            print("Error: Constraint matrix is empty!")
            exit()

        for line in scenario:
            if line[0] not in "\n":
                self.insce.append(line)

    def perform_fba(self, run):
        """ performs multiple fba """
        insce_temp = self.insce[:]
        for constraint in self.rates:
            insce_temp.append(
                "\t".join([constraint, self.rates[constraint][run]]))

        pparser = ParamParser()

        maxmin, obj_str, solver, num_iter, lb, ub = pparser.parseByHandle(
            insce_temp)

        fba_params = FbaParam(solver, maxmin, obj_str, num_iter)
        fba_params.setLinConstraints(pparser.lin_constraints)
        scenario_model = MetabolicModel()
        scenario_model.addReactions(self.model.reactions)
        scenario_model.setFiniteBounds(lb, ub, True)
        fba = FbAnalyzer()
        obj_value, solution, _ = fba.runOnModel(
            scenario_model, fba_params, rmDeadEnds=True)

        result = [obj_value]
        messages = []
        for rea in self.roi:
            try:
                result.append(solution.fluxDict[rea])
            except KeyError:
                messages.append(
                    "WARNING: %r could not be found in reactions" % rea)
                result.append(0.0)
                # self.roi.remove(rea)
        if messages and len(solution) != 0:
            print("-" * max([len(i) for i in messages]))
            print("\n".join(messages))
            print("-" * max([len(i) for i in messages]))

        if self.verb:
            print(self.names[run] + ":")
        if len(solution) != 0:
            if self.verb:
                print("Value of obj function:", obj_value)
            if self.fba_file_prefix:
                try:
                    solution.writeToFile(
                        self.fba_file_prefix + str(self.names[run]) + ".txt")
                except IOError as strerror:
                    print ("Unable to write to file %s:" %
                           os.path.basename(self.fba_file_prefix + str(self.names[run]) + ".txt"))
                    print(strerror)
                    exit()
        elif self.verb:
            print("No output written.")
        return (self.names[run],) + tuple(result)

    def run(self):
        """ performs one run at a time """
        result = []
        for run in range(self.run_count):
            res = self.perform_fba(run)
            result.append(res)
        return result

    def run_pool(self, cores):
        """ use multiprocessing for batch computation """
        pool = Pool(processes=cores)
        result = pool.map(unwrap_self_run, list(
            zip([self]*self.run_count, list(range(self.run_count)))))
        return result

    def write_solution_to_file(self, filename, solution):
        """ write solution to file """
        write_solution = [
            "\t".join(["Scenario_name", "objective"] + [i for i in self.roi])]
        for line in solution:
            write_solution.append("\t".join([str(i) for i in line]))
        with open(filename, "w") as f:
            f.write("\n".join(write_solution))


def main():
    """ parse the command line, create the mFBA class and call anylsis functions """

    usage = "Usage: %prog [options]"
    version = "multiple Flux balance analysis\n" + COPYRIGHT_VERSION_STRING
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-r", "--reactions", dest="reactionFile",
                      help="perform multiple Flux Balance Analysis on the network given "
                           "by the reaction FILE", metavar="FILE")
    parser.add_option("-p", "--parameters", dest="paramFile",
                      help="basic scenario file", metavar="FILE")
    parser.add_option("-m", "--matrix", dest="matrixFile",
                      help="matrix that contains all constraints", metavar="FILE")
    parser.add_option("-o", "--output", dest="outputFile",
                      help="write CSV result to file, tab-seperated", metavar="FILE")
    parser.add_option("-f", "--solution-files", dest="fluxFilePrefix",
                      help="write FBA solutions to files marked "
                      "with FILE-PREFIX (may contain path)",
                      metavar="FILE-PREFIX")
    parser.add_option("-v", "--verbose", action="store_true",
                      dest="verb", help="print the obj function of each FBA solution")
    parser.add_option("-c", "--cores", type="int",
                      dest="cores", help="defines how many CPU cores to use; "
                      "by default multiprocessing is disabled")
    parser.add_option("-l", "--list-of-reactions", dest="reactionsOfInterest",
                      help="a LIST of additional reactions of interest, seperated by comma; "
                      "if nothing is provided, only the obj function is written "
                      "into the result file")
    parser.set_defaults(verb=False, cores=0, reactionsOfInterest="")

    options, _ = parser.parse_args()
    parser.check_required("-r")
    parser.check_required("-p")
    parser.check_required("-m")
    parser.check_required(["-f", "-o"])

    # 1. parse reaction file
    rparser = ReactionParser()
    model = MetabolicModel()
    try:
        model.addReactionsFromFile(options.reactionFile, rparser)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in reaction file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()

    if options.reactionsOfInterest:
        reactions_of_interest = [i.strip()
                                 for i in options.reactionsOfInterest.split(",")]
    else:
        reactions_of_interest = []

    matrix = open(options.matrixFile, "r")
    matrix = matrix.readlines()
    scenario = open(options.paramFile, "r")

    mfba = MultipleFBA(model, scenario, matrix, options.fluxFilePrefix,
                       reactions_of_interest, options.verb)

    if options.cores == 0:
        result = mfba.run()

    else:
        result = mfba.run_pool(options.cores)

    if options.outputFile:
        print("write ouput to file")
        mfba.write_solution_to_file(options.outputFile, result)


if __name__ == "__main__":
    main()
